# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib
import time

from savanna.tests.integration import base
import savanna.tests.integration.configs.parameters as param


class ClusterScalingTest(base.ITestCase):

    def setUp(self):
        super(ClusterScalingTest, self).setUp()

        telnetlib.Telnet(self.host, self.port)

    def create_cluster_for_scaling(self, node_processes):
        cluster_body = self.make_cl_body_node_processes(node_processes)
        cluster_body['name'] = param.CLUSTER_NAME_SCALING

        cluster_id = self.create_cluster_and_get_id(cluster_body)

        return cluster_id

    def implement_scaling(self, cluster_id, scaling_body):
        self.put_object(self.url_cluster_with_slash, cluster_id,
                        scaling_body, 202)

        self.await_cluster_active(cluster_id)

    def implement_scaling_new_node_group_addition(self, cluster_id,
                                                  scaling_map1,
                                                  scaling_map2='',
                                                  multi_scaling=False):
        if multi_scaling:
            self.implement_scaling(cluster_id, {
                'add_node_groups': [
                {
                    'node_group_template_id': scaling_map1['ngt_id'],
                    'count': scaling_map1['node_count'],
                    'name': scaling_map1['ng_name']
                }, {
                    'node_group_template_id': scaling_map1['ngt_id'],
                    'count': scaling_map2['node_count'],
                    'name': scaling_map2['ng_name']
                }
                ]
            })

        else:
            self.implement_scaling(cluster_id, {
                'add_node_groups': [
                {
                    'node_group_template_id': scaling_map1['ngt_id'],
                    'count': scaling_map1['node_count'],
                    'name': scaling_map1['ng_name']
                }
                ]
            })

    def implement_scaling_addition_to_existing_node_group(self, cluster_id,
                                                          scaling_map1,
                                                          scaling_map2='',
                                                          multi_scaling=False):
        if multi_scaling:
            self.implement_scaling(cluster_id, {
                'resize_node_groups': [
                {
                    'name': scaling_map1['ng_name'],
                    'count': scaling_map1['node_count']

                }, {
                    'name': scaling_map2['ng_name'],
                    'count': scaling_map2['node_count']

                }
                ]
            })

        else:
            self.implement_scaling(cluster_id, {
                'resize_node_groups': [
                {
                    'name': scaling_map1['ng_name'],
                    'count': scaling_map1['node_count']

                }
                ]
            })

    def check_cluster_worker_nodes(self, cluster_id):
        ip_instances = self.get_instances_ip_and_node_processes_list(
            cluster_id)

        time.sleep(10)

        try:
            worker_map = self.get_namenode_ip_and_tt_dn_count(ip_instances)
            self.await_active_workers_for_namenode(worker_map)

            return worker_map

        except Exception as e:
            self.fail(str(e))

    def compare_worker_node_count_after_scaling(self,
                                                worker_map,
                                                worker_type,
                                                worker_node_count):
        self.assertEquals(
            worker_map[worker_type], worker_node_count,
            '%s != %s after cluster scaling!' % (worker_type, worker_type))

    def check_cluster_worker_nodes_after_scaling(self,
                                                 cluster_id,
                                                 worker_type,
                                                 scaling_worker_node_count):
        worker_map = self.check_cluster_worker_nodes(cluster_id)
        self.compare_worker_node_count_after_scaling(
            worker_map, worker_type, scaling_worker_node_count)

    def test_scaling_addition_to_existing_ng(self):
        ng_name_for_tt = 'tt'
        tt_count = 1

        ng_name_for_dn = 'dn'
        dn_count = 1
        dn_replication_factor = 3

        cluster_id = self.create_cluster_for_scaling(
            {'JT': 1, 'NN': 1, 'TT': tt_count, 'DN': dn_count})

        try:
            self.implement_scaling_addition_to_existing_node_group(
                cluster_id, {
                    'ng_name': ng_name_for_tt,
                    'node_count': tt_count + 1
                })
            self.check_cluster_worker_nodes_after_scaling(
                cluster_id, 'tasktracker_count', tt_count + 1)

            self.implement_scaling_addition_to_existing_node_group(
                cluster_id, {
                    'ng_name': ng_name_for_dn,
                    'node_count': dn_count + dn_replication_factor
                })
            self.check_cluster_worker_nodes_after_scaling(
                cluster_id, 'datanode_count', dn_count + dn_replication_factor)

            multi_scaling = True
            self.implement_scaling_addition_to_existing_node_group(
                cluster_id, {
                    'ng_name': ng_name_for_tt,
                    'node_count': 0
                },
                {
                    'ng_name': ng_name_for_dn,
                    'node_count': dn_replication_factor
                },
                multi_scaling)
            self.check_cluster_worker_nodes_after_scaling(
                cluster_id, 'tasktracker_count', 0)
            self.check_cluster_worker_nodes_after_scaling(
                cluster_id, 'datanode_count', dn_replication_factor)

        except Exception as e:
            self.fail(str(e))

        finally:
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)

    def test_scaling_new_node_group_addition(self):
        ng_name_for_tt = 'ng-tt'
        added_tt_count = 2

        ng_name_for_dn = 'ng-dn'
        added_dn_count = 2
        dn_replication_factor = 3

        cluster_id = self.create_cluster_for_scaling(
            {'JT+NN': 1, 'TT+DN': 1})

        self.create_node_group_templates()

        try:
            self.implement_scaling_new_node_group_addition(
                cluster_id, {
                    'ngt_id': self.id_tt,
                    'node_count': added_tt_count,
                    'ng_name': ng_name_for_tt,

                })
            self.check_cluster_worker_nodes_after_scaling(
                cluster_id, 'tasktracker_count', added_tt_count + 1)

            self.implement_scaling_new_node_group_addition(
                cluster_id, {
                    'ngt_id': self.id_dn,
                    'node_count': added_dn_count + dn_replication_factor,
                    'ng_name': ng_name_for_dn,

                })
            self.check_cluster_worker_nodes_after_scaling(
                cluster_id, 'datanode_count',
                1 + added_dn_count + dn_replication_factor)

            multi_scaling = True
            self.implement_scaling_addition_to_existing_node_group(
                cluster_id, {
                    'ngt_id': self.id_tt,
                    'node_count': 0,
                    'ng_name': ng_name_for_tt,

                }, {
                    'ngt_id': self.id_dn,
                    'node_count': dn_replication_factor,
                    'ng_name': ng_name_for_dn,

                }, multi_scaling)
            self.check_cluster_worker_nodes_after_scaling(
                cluster_id, 'tasktracker_count', 1)
            self.check_cluster_worker_nodes_after_scaling(
                cluster_id, 'datanode_count', 1 + dn_replication_factor)

        except Exception as e:
            self.fail(str(e))

        finally:
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)
            self.delete_node_group_templates()



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2012 Red Hat, Inc.
# All Rights Reserved.
# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
gettext for openstack-common modules.

Usual usage in an openstack.common module:

    from savanna.openstack.common.gettextutils import _
"""

import copy
import gettext
import logging.handlers
import os
import UserString

_localedir = os.environ.get('savanna'.upper() + '_LOCALEDIR')
_t = gettext.translation('savanna', localedir=_localedir, fallback=True)


def _(msg):
    return _t.ugettext(msg)


def install(domain):
    """Install a _() function using the given translation domain.

    Given a translation domain, install a _() function using gettext's
    install() function.

    The main difference from gettext.install() is that we allow
    overriding the default localedir (e.g. /usr/share/locale) using
    a translation-domain-specific environment variable (e.g.
    NOVA_LOCALEDIR).
    """
    gettext.install(domain,
                    localedir=os.environ.get(domain.upper() + '_LOCALEDIR'),
                    unicode=True)


"""
Lazy gettext functionality.

The following is an attempt to introduce a deferred way
to do translations on messages in OpenStack. We attempt to
override the standard _() function and % (format string) operation
to build Message objects that can later be translated when we have
more information. Also included is an example LogHandler that
translates Messages to an associated locale, effectively allowing
many logs, each with their own locale.
"""


def get_lazy_gettext(domain):
    """Assemble and return a lazy gettext function for a given domain.

    Factory method for a project/module to get a lazy gettext function
    for its own translation domain (i.e. nova, glance, cinder, etc.)
    """

    def _lazy_gettext(msg):
        """Create and return a Message object.

        Message encapsulates a string so that we can translate it later when
        needed.
        """
        return Message(msg, domain)

    return _lazy_gettext


class Message(UserString.UserString, object):
    """Class used to encapsulate translatable messages."""
    def __init__(self, msg, domain):
        # _msg is the gettext msgid and should never change
        self._msg = msg
        self._left_extra_msg = ''
        self._right_extra_msg = ''
        self.params = None
        self.locale = None
        self.domain = domain

    @property
    def data(self):
        # NOTE(mrodden): this should always resolve to a unicode string
        # that best represents the state of the message currently

        localedir = os.environ.get(self.domain.upper() + '_LOCALEDIR')
        if self.locale:
            lang = gettext.translation(self.domain,
                                       localedir=localedir,
                                       languages=[self.locale],
                                       fallback=True)
        else:
            # use system locale for translations
            lang = gettext.translation(self.domain,
                                       localedir=localedir,
                                       fallback=True)

        full_msg = (self._left_extra_msg +
                    lang.ugettext(self._msg) +
                    self._right_extra_msg)

        if self.params is not None:
            full_msg = full_msg % self.params

        return unicode(full_msg)

    def _save_parameters(self, other):
        # we check for None later to see if
        # we actually have parameters to inject,
        # so encapsulate if our parameter is actually None
        if other is None:
            self.params = (other, )
        else:
            self.params = copy.deepcopy(other)

        return self

    # overrides to be more string-like
    def __unicode__(self):
        return self.data

    def __str__(self):
        return self.data.encode('utf-8')

    def __getstate__(self):
        to_copy = ['_msg', '_right_extra_msg', '_left_extra_msg',
                   'domain', 'params', 'locale']
        new_dict = self.__dict__.fromkeys(to_copy)
        for attr in to_copy:
            new_dict[attr] = copy.deepcopy(self.__dict__[attr])

        return new_dict

    def __setstate__(self, state):
        for (k, v) in state.items():
            setattr(self, k, v)

    # operator overloads
    def __add__(self, other):
        copied = copy.deepcopy(self)
        copied._right_extra_msg += other.__str__()
        return copied

    def __radd__(self, other):
        copied = copy.deepcopy(self)
        copied._left_extra_msg += other.__str__()
        return copied

    def __mod__(self, other):
        # do a format string to catch and raise
        # any possible KeyErrors from missing parameters
        self.data % other
        copied = copy.deepcopy(self)
        return copied._save_parameters(other)

    def __mul__(self, other):
        return self.data * other

    def __rmul__(self, other):
        return other * self.data

    def __getitem__(self, key):
        return self.data[key]

    def __getslice__(self, start, end):
        return self.data.__getslice__(start, end)

    def __getattribute__(self, name):
        # NOTE(mrodden): handle lossy operations that we can't deal with yet
        # These override the UserString implementation, since UserString
        # uses our __class__ attribute to try and build a new message
        # after running the inner data string through the operation.
        # At that point, we have lost the gettext message id and can just
        # safely resolve to a string instead.
        ops = ['capitalize', 'center', 'decode', 'encode',
               'expandtabs', 'ljust', 'lstrip', 'replace', 'rjust', 'rstrip',
               'strip', 'swapcase', 'title', 'translate', 'upper', 'zfill']
        if name in ops:
            return getattr(self.data, name)
        else:
            return UserString.UserString.__getattribute__(self, name)


class LocaleHandler(logging.Handler):
    """Handler that can have a locale associated to translate Messages.

    A quick example of how to utilize the Message class above.
    LocaleHandler takes a locale and a target logging.Handler object
    to forward LogRecord objects to after translating the internal Message.
    """

    def __init__(self, locale, target):
        """Initialize a LocaleHandler

        :param locale: locale to use for translating messages
        :param target: logging.Handler object to forward
                       LogRecord objects to after translation
        """
        logging.Handler.__init__(self)
        self.locale = locale
        self.target = target

    def emit(self, record):
        if isinstance(record.msg, Message):
            # set the locale and resolve to a string
            record.msg.locale = self.locale

        self.target.emit(record)



# Copyright (c) 2013 Hortonworks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna.plugins.hdp import baseprocessor as b


class BlueprintProcessor(b.BaseProcessor):

    def __init__(self, blueprint):
        self.blueprint = blueprint

    def process_user_inputs(self, user_inputs):
        for ui in user_inputs:
            configurations = self.blueprint['configurations']
            properties_dict = self._find_blueprint_section(configurations,
                                                           'name',
                                                           ui.config.file)

            if properties_dict is None:
                properties_dict = {'name': ui.config.file, 'properties': []}
                configurations.append(properties_dict)

            # need to look for property with
            # the same name
            properties = properties_dict['properties']
            prop = self._find_blueprint_section(properties, 'name',
                                                ui.config.name)
            if prop is not None:
                # time to change the value
                prop['value'] = ui.value
            else:
                prop = {'name': ui.config.name, 'value': ui.value}
                properties.append(prop)

    def process_node_groups(self, node_groups):
        # we're overwriting existing settings
        if node_groups:
            for node_group in node_groups:
                # we're either replacing existing settings for the group or
                # creating a new one.  either way
                if node_group.node_processes:
                    host_role_mapping = {u'name': unicode(node_group.name),
                                         u'components': [], u'hosts': []}
                    for process in node_group.node_processes:
                        host_role_mapping[u'components'].append(
                            {u'name': unicode(process.upper())})
                    if node_group.count > 1:
                        host_role_mapping[u'hosts'].append(
                            {u'cardinality': u'1+'})
                    else:
                        host_role_mapping[u'hosts'].append(
                            {u'cardinality': u'1'})

                    existing_mapping = next(
                        (item for item in self.blueprint['host_role_mappings']
                         if item['name'] == host_role_mapping['name']), None)

                    if (existing_mapping is not None):
                        mappings__index = self.blueprint[
                            'host_role_mappings'].index(existing_mapping)
                        self.blueprint['host_role_mappings'][
                            mappings__index] = host_role_mapping
                    else:
                        self.blueprint['host_role_mappings'].append(
                            host_role_mapping)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest2

from savanna.db import models as m
from savanna.plugins.vanilla import utils as u


class VanillaUtilsTest(unittest2.TestCase):
    def setUp(self):
        self.c1 = m.Cluster("cluster1", "tenant1", "vanilla", "1.1.2")
        self.ng1 = m.NodeGroup("master", "f1", ["jt", "nn"], 1)
        self.ng2 = m.NodeGroup("workers", "f1", ["tt", "dn"], 3)
        self.ng3 = m.NodeGroup("sn", "f1", ["dn"], 1)
        self.c1.node_groups = [self.ng1, self.ng2, self.ng3]

        self.ng1.instances = [m.Instance("ng1", "i1", "master")]
        self.ng2.instances = [m.Instance("ng2", "i2", "worker1"),
                              m.Instance("ng2", "i3", "worker2"),
                              m.Instance("ng2", "i4", "worker3")]
        self.ng3.instances = [m.Instance("ng3", "i5", "sn")]

    def test_get_node_groups(self):
        self.assertListEqual(u.get_node_groups(self.c1), self.c1.node_groups)
        self.assertListEqual(u.get_node_groups(self.c1, ["wrong-process"]), [])
        self.assertListEqual(u.get_node_groups(self.c1, ['dn', 'tt']),
                             [self.ng2])
        self.assertListEqual(u.get_node_groups(self.c1, 'dn'),
                             [self.ng2, self.ng3])
        self.assertListEqual(u.get_node_groups(self.c1, ['dn']),
                             [self.ng2, self.ng3])
        self.assertListEqual(u.get_node_groups(self.c1, ['jt', 'tt']), [])

    def test_get_instances(self):
        self.assertEquals(len(u.get_instances(self.c1)), 5)
        self.assertListEqual(u.get_instances(self.c1, 'wrong-process'), [])
        self.assertListEqual(u.get_instances(self.c1, 'nn'),
                             self.ng1.instances)
        self.assertListEqual(u.get_instances(self.c1, 'dn'),
                             self.ng2.instances + self.ng3.instances)

    def test_generate_lines_from_list(self):
        self.assertEquals(u.generate_host_names(self.ng2.instances),
                          "worker1\nworker2\nworker3")
        self.assertEquals(u.generate_host_names([]), "")



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest2

from savanna.utils import crypto as c


class CryptoTest(unittest2.TestCase):
    def test_generate_private_key(self):
        pk = c.generate_private_key()

        self.assertIsNotNone(pk)
        self.assertIn('-----BEGIN RSA PRIVATE KEY-----', pk)
        self.assertIn('-----END RSA PRIVATE KEY-----', pk)

    def test_to_paramiko_private_key(self):
        pk_str = c.generate_private_key()
        pk = c.to_paramiko_private_key(pk_str)

        self.assertIsNotNone(pk)
        self.assertEqual(2048, pk.size)
        self.assertEqual('ssh-rsa', pk.get_name())

    def test_private_key_to_public_key(self):
        key = c.private_key_to_public_key(c.generate_private_key())

        self.assertIsNotNone(key)
        self.assertIn('ssh-rsa', key)



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Base class(es) for WSGI Middleware."""

import webob.dec


class Middleware(object):
    """Base WSGI middleware wrapper.

    These classes require an application to be initialized that will be called
    next.  By default the middleware will simply call its wrapped app, or you
    can override __call__ to customize its behavior.
    """

    @classmethod
    def factory(cls, global_conf, **local_conf):
        """Factory method for paste.deploy."""

        def filter(app):
            return cls(app)

        return filter

    def __init__(self, application):
        self.application = application

    def process_request(self, req):
        """Called on each request.

        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.
        """
        return None

    def process_response(self, response):
        """Do whatever you'd like to the response."""
        return response

    @webob.dec.wsgify
    def __call__(self, req):
        response = self.process_request(req)
        if response:
            return response
        response = req.get_response(self.application)
        return self.process_response(response)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from Crypto.PublicKey import RSA
from Crypto import Random
import paramiko
import six


def generate_private_key(length=2048):
    """Generate RSA private key (str) with the specified length."""
    rsa = RSA.generate(length, Random.new().read)
    return rsa.exportKey('PEM')


def to_paramiko_private_key(pkey):
    """Convert private key (str) to paramiko-specific RSAKey object."""
    return paramiko.RSAKey(file_obj=six.StringIO(pkey))


def private_key_to_public_key(key):
    """Convert private key (str) to public key (str)."""
    return RSA.importKey(key).exportKey('OpenSSH')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import novaclient.exceptions as nova_ex
import unittest2

import savanna.db.models as m
from savanna.plugins.vanilla import plugin
import savanna.service.validation as v

_types_checks = {
    "string": [1, (), {}],
    "integer": ["a", (), {}],
    "uuid": ["z550e8400-e29b-41d4-a716-446655440000", 1, "a", (), {}],
    "array": [{}, 'a', 1]
}


def _update_data(data, update):
    data.update(update)
    return data


def _get_plugins():
    vanilla = plugin.VanillaProvider
    vanilla.name = 'vanilla'
    return [vanilla]


def _get_plugin(name):
    if name == 'vanilla':
        vanilla = plugin.VanillaProvider
        vanilla.name = 'vanilla'
        return vanilla
    return None


def _get_keypair(name):
    if name is not "test_keypair":
        raise nova_ex.NotFound("")


def _get_flavor(flavor_id):
    if flavor_id is not "42":
        raise nova_ex.NotFound("")


def start_patch():
    get_clusters_p = mock.patch("savanna.service.api.get_clusters")
    get_cluster_p = mock.patch("savanna.service.api.get_cluster")
    get_ng_templates_p = \
        mock.patch("savanna.service.api.get_node_group_templates")
    get_ng_template_p = \
        mock.patch("savanna.service.api.get_node_group_template")
    get_plugins_p = mock.patch("savanna.service.api.get_plugins")
    get_plugin_p = \
        mock.patch("savanna.plugins.base.PluginManager.get_plugin")
    get_cl_templates_p = \
        mock.patch("savanna.service.api.get_cluster_templates")
    get_cl_template_p = \
        mock.patch("savanna.service.api.get_cluster_template")
    nova_p = mock.patch("savanna.utils.openstack.nova.client")
    keystone_p = mock.patch("savanna.utils.openstack.keystone.client")
    get_image_p = mock.patch("savanna.service.api.get_image")

    get_image = get_image_p.start()
    get_clusters = get_clusters_p.start()
    get_cluster = get_cluster_p.start()
    get_ng_templates = get_ng_templates_p.start()
    get_ng_template = get_ng_template_p.start()
    get_plugins = get_plugins_p.start()
    get_plugin = get_plugin_p.start()
    get_cl_templates = get_cl_templates_p.start()
    get_cl_template_p.start()

    nova = nova_p.start()
    keystone = keystone_p.start()

    get_cl_templates.return_value = []

    nova().flavors.get.side_effect = _get_flavor
    nova().keypairs.get.side_effect = _get_keypair

    class Service:
        @property
        def name(self):
            return 'cinder'

    def _services_list():
        return [Service()]

    keystone().services.list.side_effect = _services_list

    class Image:
        def __init__(self, name='test'):
            self.name = name

        @property
        def id(self):
            if self.name == 'test':
                return '550e8400-e29b-41d4-a716-446655440000'
            else:
                return '813fe450-40d2-4acc-ade5-ea753a1bd5bc'

        @property
        def tags(self):
            if self.name == 'test':
                return ['vanilla', '1.1.2']
            else:
                return ['wrong_tag']

    def _get_image(id):
        if id == '550e8400-e29b-41d4-a716-446655440000':
            return Image()
        else:
            return Image('wrong_test')

    get_image.side_effect = _get_image
    nova().images.list_registered.return_value = [Image(),
                                                  Image(name='wrong_name')]
    cluster = m.Cluster('test', 't', 'vanilla', '1.2.2')
    cluster.id = 1
    cluster.status = 'Active'
    cluster.node_groups = [m.NodeGroup('ng', '42', ['namenode'], 1)]
    # stub clusters list
    get_clusters.return_value = [cluster]
    get_cluster.return_value = cluster
    # stub node templates
    get_ng_templates.return_value = [m.NodeGroupTemplate('test', 't', '42',
                                                         'vanilla', '1.2.2',
                                                         ['namenode'])]

    get_cl_templates.return_value = [m.ClusterTemplate('test', 't',
                                                       'vanilla', '1.2.2')]

    vanilla = plugin.VanillaProvider()
    vanilla.name = 'vanilla'
    get_plugins.return_value = [vanilla]

    def _get_ng_template(name):
        for template in get_ng_templates():
            if template.name == name:
                return template
        return None

    def _get_plugin(name):
        if name == 'vanilla':
            return vanilla
        return None

    get_plugin.side_effect = _get_plugin
    get_ng_template.side_effect = _get_ng_template
    # request data to validate
    patchers = (get_clusters_p, get_ng_templates_p, get_ng_template_p,
                get_plugins_p, get_plugin_p,
                get_cl_template_p, get_cl_templates_p, nova_p, keystone_p,
                get_image_p)
    return patchers


def stop_patch(patchers):
    for patcher in patchers:
        patcher.stop()


class ValidationTestCase(unittest2.TestCase):
    def setUp(self):
        self._create_object_fun = None
        self.scheme = None

    def tearDown(self):
        self._create_object_fun = None

    def _assert_calls(self, mock, call_info):
        if not call_info:
            self.assertEqual(mock.call_count, 0)
        else:
            self.assertEqual(mock.call_count, call_info[0])
            self.assertEqual(mock.call_args[0][0].code, call_info[1])
            self.assertEqual(mock.call_args[0][0].message, call_info[2])

    @mock.patch("savanna.utils.api.request_data")
    @mock.patch("savanna.utils.api.bad_request")
    def _assert_create_object_validation(
            self, bad_req=None, request_data=None,
            data=None, bad_req_i=None):

        request_data.return_value = data
        # mock function that should be validated
        patchers = start_patch()
        m_func = mock.Mock()
        m_func.__name__ = "m_func"
        v.validate(self.scheme, self._create_object_fun)(m_func)(data=data)

        self.assertEqual(request_data.call_count, 1)
        self._assert_calls(bad_req, bad_req_i)
        stop_patch(patchers)

    def _assert_object_name_validation(self, data, name_type):

        data.update({'name': None})
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"None is not of type 'string'")
        )
        data.update({'name': ""})
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'' is too short")
        )
        data.update({'name': ('a' * 51)})
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'%s' is too long" % ('a' * 51))
        )
        data.update({'name': 'a-!'})
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'a-!' is not a '%s'" % name_type)
        )
        data.update({'name': 'aa_'})
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'aa_' is not a '%s'" % name_type)
        )
        # For cluster templates validating
        if name_type == 'valid_name':
            data.update({'name': '1-a'})
            self._assert_create_object_validation(
                data=data,
                bad_req_i=(1, "VALIDATION_ERROR",
                           u"'1-a' is not a '%s'" % name_type)
            )

    def _assert_types(self, default_data):
        for p_name in self.scheme['properties']:
            prop = self.scheme['properties'][p_name]
            if prop["type"] in _types_checks:
                for type_ex in _types_checks[prop["type"]]:
                    data = default_data.copy()
                    value = type_ex
                    value_str = str(value)
                    if isinstance(value, str):
                        value_str = "'%s'" % value_str
                    data.update({p_name: value})
                    self._assert_create_object_validation(
                        data=data,
                        bad_req_i=(1, 'VALIDATION_ERROR',
                                   u"%s is not of type '%s'"
                                   % (value_str, prop["type"]))
                    )

    def _assert_cluster_configs_validation(self):
        data = {
            'name': 'test-cluster',
            'plugin_name': 'vanilla',
            'hadoop_version': '1.1.2',
            'cluster_configs': {
                'HDFS': {
                    u'hadoop.tmp.dir': '/temp/'
                }
            }
        }
        self._assert_create_object_validation(
            data=_update_data(data, {
                'cluster_configs': {
                    'wrong_target': {
                        u'hadoop.tmp.dir': '/temp/'
                    }
                }}),
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Plugin doesn't contain applicable "
                       "target 'wrong_target'")
        )
        self._assert_create_object_validation(
            data=_update_data(data, {
                'cluster_configs': {
                    'HDFS': {
                        u's': '/temp/'
                    }
                }
            }),
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Plugin's applicable target 'HDFS' doesn't "
                       "contain config with name 's'")
        )

    def _assert_cluster_default_image_tags_validation(self):
        data = {
            'name': 'test-cluster',
            'plugin_name': 'vanilla',
            'hadoop_version': '1.1.2',
            'default_image_id': '550e8400-e29b-41d4-a716-446655440000'
        }
        self._assert_create_object_validation(data=data)
        data = {
            'name': 'test-cluster',
            'plugin_name': 'vanilla',
            'hadoop_version': '1.1.2',
            'default_image_id': '813fe450-40d2-4acc-ade5-ea753a1bd5bc'
        }
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Tags of requested image "
                       "'813fe450-40d2-4acc-ade5-ea753a1bd5bc' "
                       "don't contain required tags "
                       "['vanilla', '1.1.2']"))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import savanna.service.validations.base as b
import savanna.service.validations.cluster_templates as cl_tmpl


def _build_cluster_schema():
    cluster_schema = copy.deepcopy(cl_tmpl.CLUSTER_TEMPLATE_SCHEMA)
    cluster_schema['properties']['name']['format'] = "hostname"
    cluster_schema['properties'].update({
        "user_keypair_id": {
            "type": "string",
            "format": "valid_name",
        },
        "cluster_template_id": {
            "type": "string",
            "format": "uuid",
        }})
    return cluster_schema

CLUSTER_SCHEMA = _build_cluster_schema()


def check_cluster_create(data, **kwargs):
    b.check_cluster_unique_name(data['name'])
    b.check_plugin_name_exists(data['plugin_name'])
    b.check_plugin_supports_version(data['plugin_name'],
                                    data['hadoop_version'])
    if data.get('cluster_template_id'):
        b.check_cluster_template_exists(data['cluster_template_id'])

    if data.get('user_keypair_id'):
        b.check_keypair_exists(data['user_keypair_id'])

    if data.get('default_image_id'):
        b.check_image_registered(data['default_image_id'])
        b.check_required_image_tags(data['plugin_name'],
                                    data['hadoop_version'],
                                    data['default_image_id'])

    b.check_all_configurations(data)

    if data.get('anti_affinity'):
        b.check_node_processes(data['plugin_name'], data['hadoop_version'],
                               data['anti_affinity'])



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna.service import api
from savanna.service.validations import node_group_templates as nt
from savanna.tests.unit.service.validation import utils as u


class TestNGTemplateCreateValidation(u.ValidationTestCase):
    def setUp(self):
        self._create_object_fun = nt.check_node_group_template_create
        self.scheme = nt.NODE_GROUP_TEMPLATE_SCHEMA
        api.plugin_base.setup_plugins()

    def test_node_groups_create_required(self):
        self._assert_create_object_validation(
            data={
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'name' is a required property")
        )
        self._assert_create_object_validation(
            data={
                'name': 'a'
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'flavor_id' is a required property")
        )
        self._assert_create_object_validation(
            data={
                'name': 'a',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'hadoop_version' is a required property")
        )
        self._assert_create_object_validation(
            data={
                'name': 'a',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2'
            },
            bad_req_i=(1, "VALIDATION_ERROR",
                       u"'node_processes' is a required property")
        )
        self._assert_create_object_validation(
            data={
                'name': "a",
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': []
            },
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u'[] is too short')
        )

    def test_ng_template_create_v_names(self):
        data = {
            'name': 'a',
            'flavor_id': '42',
            'plugin_name': 'vanilla',
            'hadoop_version': '1.1.2',
            'node_processes': ['namenode']
        }
        self._assert_object_name_validation(data, "valid_name")

    def test_ng_template_create_v_node_processes(self):
        self._assert_create_object_validation(
            data={
                'name': "a",
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ["namenode", "namenode"]
            },
            bad_req_i=(1, 'INVALID_REFERENCE',
                       'Duplicates in node processes have been detected')
        )
        self._assert_create_object_validation(
            data={
                'name': 'a',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ['wrong_process']
            },
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Plugin supports the following node procesess: "
                       "['namenode', 'datanode', 'secondarynamenode', "
                       "'tasktracker', 'jobtracker']")
        )

    def test_ng_template_create_v_right(self):
        self._assert_create_object_validation(
            data={
                'name': 'a',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ['namenode',
                                   'datanode',
                                   'secondarynamenode',
                                   'tasktracker',
                                   'jobtracker'],
                'node_configs': {
                    'HDFS': {
                        u'hadoop.tmp.dir': '/temp/'
                    }
                },
                'image_id': '550e8400-e29b-41d4-a716-446655440000',
                'volumes_per_node': 2,
                'volumes_size': 10,
                'description': 'test node template'

            }
        )

    def test_ng_template_create_v_minimum_ints(self):
        self._assert_create_object_validation(
            data={
                'name': 'a',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ['wrong_process'],
                'volumes_per_node': -1
            },
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u'-1.0 is less than the minimum of 0')
        )
        self._assert_create_object_validation(
            data={
                'name': 'a',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ['wrong_process'],
                'volumes_size': 0
            },
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u'0.0 is less than the minimum of 1')
        )

    def test_ng_template_create_v_types(self):
        default_data = {
            'name': 'a', 'flavor_id': '42',
            'plugin_name': 'vanilla',
            'hadoop_version': '1.1.2',
            'node_processes': ['namenode']
        }
        self._assert_types(default_data)

    def test_ng_template_create_v_unique_ng(self):
        data = {
            'name': 'test',
            'flavor_id': '42',
            'plugin_name': 'vanilla',
            'hadoop_version': '1.1.2',
            'node_processes': ['namenode']}
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, 'NAME_ALREADY_EXISTS',
                       "NodeGroup template with name 'test' already exists")
        )

    def test_ng_template_create_v_flavor_exists(self):
        self._assert_create_object_validation(
            data={
                'name': 'test-ng',
                'flavor_id': '1',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ['namenode']
            },
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Requested flavor '1' not found")
        )

    def test_ng_template_create_v_ng_configs(self):
        self._assert_create_object_validation(
            data={
                'name': 'test-ng',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ['namenode'],
                'node_configs': {
                    'wrong_target': {
                        u'hadoop.tmp.dir': '/temp/'
                    }
                }},
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Plugin doesn't contain applicable "
                       "target 'wrong_target'")
        )
        self._assert_create_object_validation(
            data={
                'name': 'test-ng',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ['namenode'],
                'node_configs': {
                    'HDFS': {
                        's': 'a'
                    }
                }
            },
            bad_req_i=(1, 'INVALID_REFERENCE',
                       "Plugin's applicable target 'HDFS' doesn't "
                       "contain config with name 's'")
        )

    def test_ng_template_cinder(self):
        self._assert_create_object_validation(
            data={
                'name': 'a',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ['wrong_process'],
                'volumes_per_node': -1
            },
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u'-1.0 is less than the minimum of 0')
        )
        self._assert_create_object_validation(
            data={
                'name': 'a',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ['wrong_process'],
                'volumes_size': 0
            },
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u'0.0 is less than the minimum of 1')
        )
        self._assert_create_object_validation(
            data={
                'name': 'a',
                'flavor_id': '42',
                'plugin_name': 'vanilla',
                'hadoop_version': '1.1.2',
                'node_processes': ['datanode', 'tasktracker'],
                'volumes_per_node': 1,
                'volumes_size': 1,
                'volume_mount_prefix': '/mnt/volume'
            }
        )
        data = {
            'name': 'a',
            'flavor_id': '42',
            'plugin_name': 'vanilla',
            'hadoop_version': '1.1.2',
            'node_processes': ['datanode', 'tasktracker'],
            'volumes_per_node': 1,
            'volumes_size': 1,
            'volume_mount_prefix': 'qwerty'
        }
        self._assert_create_object_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR', "'qwerty' is not a 'posix_path'")
        )



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from cinderclient.v1 import volumes as v
from savanna.db import models as m
from savanna.service import volumes
from savanna.tests.unit import base as models_test_base


class TestAttachVolume(models_test_base.DbTestCase):
    @mock.patch(
        'savanna.utils.remote.BulkInstanceInteropHelper.execute_command')
    def test_mount_volume(self, p_ex_cmd):
        instance = m.Instance(None, None, None)

        p_ex_cmd.return_value = (0, None)
        self.assertIsNone(volumes._mount_volume(instance, '123', '456'))
        self.assertEqual(p_ex_cmd.call_count, 3)
        p_ex_cmd.reset_mock()

        p_ex_cmd.return_value = (1, None)
        self.assertRaises(RuntimeError, volumes._mount_volume,
                          instance, '123', '456')
        self.assertEqual(p_ex_cmd.call_count, 3)

    @mock.patch('cinderclient.v1.volumes.Volume.delete')
    @mock.patch('cinderclient.v1.volumes.Volume.detach')
    @mock.patch('savanna.utils.openstack.cinder.get_volume')
    def test_detach_volumes(self, p_get_volume, p_detach, p_delete):
        instance = m.Instance(None, None, None)
        instance.volumes.append("123")

        p_get_volume.return_value = v.Volume(None, {'id': '123'})
        p_detach.return_value = None
        p_delete.return_value = None
        self.assertIsNone(
            volumes.detach_from_instances([instance]))

        ng = m.NodeGroup(None, None, None, None)
        ng.instances.append(instance)
        cluster = m.Cluster(None, None, None, None)
        cluster.node_groups.append(ng)
        p_delete.side_effect = RuntimeError
        self.assertRaises(RuntimeError, volumes.detach, cluster)

    @mock.patch('savanna.utils.remote.InstanceInteropHelper.execute_command')
    def test_get_free_device_path(self, p_ex_cmd):
        instance = m.Instance(None, None, None)

        p_ex_cmd.return_value = (1, None)
        self.assertRaises(RuntimeError, volumes._get_free_device_path,
                          instance)

        stdout = """major minor  #blocks  name

   8        0  488386584 vda
   8        1     102400 vda1"""

        p_ex_cmd.return_value = (0, stdout)
        self.assertEqual(volumes._get_free_device_path(instance), '/dev/vdb')

        stdout = "major minor  #blocks  name\n"
        for idx in range(0, 26):
            line = "   8        0  488386584 vd" + chr(ord('a') + idx) + '\n'
            stdout += line

        p_ex_cmd.return_value = (0, stdout)
        self.assertRaises(RuntimeError, volumes._get_free_device_path,
                          instance)

    @mock.patch('savanna.service.volumes._mount_volume')
    @mock.patch('savanna.service.volumes._await_attach_volume')
    @mock.patch('savanna.service.volumes._create_attach_volume')
    @mock.patch('savanna.service.volumes._get_free_device_path')
    def test_attach(self, p_dev_path, p_create_attach_vol,
                    p_await, p_mount):
        p_dev_path.return_value = '123'
        p_create_attach_vol.return_value = None
        p_await.return_value = None
        p_mount.return_value = None
        ng = m.NodeGroup(None, None, None, None, volumes_per_node=2,
                         volumes_size=2)
        instance1 = m.Instance(None, None, '123')
        instance2 = m.Instance(None, None, '456')
        instance1.node_group_id = ng.id
        instance2.node_group_id = ng.id
        ng.instances.append(instance1)
        ng.instances.append(instance2)
        cluster = m.Cluster(None, None, None, None)
        cluster.node_groups.append(ng)

        volumes.attach(cluster)
        self.assertEqual(p_create_attach_vol.call_count, 4)
        self.assertEqual(p_await.call_count, 4)
        self.assertEqual(p_mount.call_count, 4)
        self.assertEqual(p_dev_path.call_count, 4)

        p_create_attach_vol.reset_mock()
        p_await.reset_mock()
        p_mount.reset_mock()
        p_dev_path.reset_mock()

        instances = [instance1, instance2]
        volumes.attach_to_instances(instances)

        self.assertEqual(p_create_attach_vol.call_count, 4)
        self.assertEqual(p_await.call_count, 4)
        self.assertEqual(p_mount.call_count, 4)
        self.assertEqual(p_dev_path.call_count, 4)

    @mock.patch('savanna.context.sleep')
    @mock.patch('savanna.service.volumes._get_device_paths')
    def test_await_attach_volume(self, dev_paths, p_sleep):
        dev_paths.return_value = ['/dev/vda', '/dev/vdb']
        p_sleep.return_value = None
        instance = m.Instance(None, None, None)
        self.assertIsNone(volumes._await_attach_volume(instance, '/dev/vdb'))
        self.assertRaises(RuntimeError, volumes._await_attach_volume,
                          instance, '/dev/vdc')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import jsonschema
import six

from savanna.openstack.common import uuidutils


@jsonschema.FormatChecker.cls_checks('valid_name')
def validate_name_format(entry):
    res = re.match(r"^(([a-zA-Z]|[a-zA-Z][a-zA-Z0-9\-_]"
                   r"*[a-zA-Z0-9])\.)*([A-Za-z]|[A-Za-z]"
                   r"[A-Za-z0-9\-_]*[A-Za-z0-9])$", entry)
    return res is not None


@jsonschema.FormatChecker.cls_checks('valid_tag')
def validate_valid_tag_format(entry):
    res = re.match(r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-_]"
                   r"*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9]"
                   r"[A-Za-z0-9\-_]*[A-Za-z0-9])$", entry)
    return res is not None


@jsonschema.FormatChecker.cls_checks('uuid')
def validate_uuid_format(entry):
    return uuidutils.is_uuid_like(entry)


@jsonschema.FormatChecker.cls_checks('posix_path')
def validate_posix_path(entry):
    res = re.match("^(/([A-Z]|[a-z]|[0-9]|\-|_)+)*$", entry)
    return res is not None


class ConfigTypeMeta(type):
    def __instancecheck__(cls, instance):
        # configs should be dict
        if not isinstance(instance, dict):
            return False

        # check dict content
        for applicable_target, configs in six.iteritems(instance):
            # upper-level dict keys (applicable targets) should be strings
            if not isinstance(applicable_target, six.string_types):
                return False

            # upper-level dict values should be dicts
            if not isinstance(configs, dict):
                return False

            # check internal dict content
            for config_name, config_value in six.iteritems(configs):
                # internal dict keys should be strings
                if not isinstance(config_name, six.string_types):
                    return False

                # internal dict values should be strings or integers or bools
                if not isinstance(config_value,
                                  (six.string_types, six.integer_types)):
                    return False

        return True


class ConfigsType(dict):
    __metaclass__ = ConfigTypeMeta


class FlavorTypeMeta(type):
    def __instancecheck__(cls, instance):
        try:
            int(instance)
        except (ValueError, TypeError):
            return (isinstance(instance, six.string_types)
                    and uuidutils.is_uuid_like(instance))
        return (isinstance(instance, six.integer_types + six.string_types)
                and type(instance) != bool)


class FlavorType(object):
    __metaclass__ = FlavorTypeMeta


class ApiValidator(jsonschema.Draft4Validator):
    def __init__(self, schema):
        format_checker = jsonschema.FormatChecker()
        super(ApiValidator, self).__init__(
            schema, format_checker=format_checker, types={
                "configs": ConfigsType,
                "flavor": FlavorType,
            })



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib

from savanna.tests.integration import base


class NodeGroupTemplatesCrudTest(base.ITestCase):

    def setUp(self):
        super(NodeGroupTemplatesCrudTest, self).setUp()
        telnetlib.Telnet(self.host, self.port)

    def test_crud_ngt_nn(self):
        """This test checks node group template creation with
        configuration | NN |.
        """
        body_nn = self.make_node_group_template('master-nn', 'qa probe', 'NN')
        self.crud_object(body_nn, self.url_ngt)

    def test_crud_ngt_jt(self):
        """This test checks node group template creation with
        configuration | JT |.
        """
        body_jt = self.make_node_group_template('master-jt', 'qa probe', 'JT')
        self.crud_object(body_jt, self.url_ngt)

    def test_crud_ngt_tt(self):
        """This test checks node group template creation with
        configuration | TT |.
        """
        body_tt = self.make_node_group_template('worker-tt', 'qa probe', 'TT')
        self.crud_object(body_tt, self.url_ngt)

    def test_crud_ngt_dn(self):
        """This test checks node group template creation with
        configuration | DN |.
        """
        body_dn = self.make_node_group_template('worker-dn', 'qa probe', 'DN')
        self.crud_object(body_dn, self.url_ngt)

    def test_crud_ngt_nn_jt(self):
        """This test checks node group template creation with
        configuration | JT + NN |.
        """
        body_nn_jt = self.make_node_group_template('master-nn-jt', 'qa probe',
                                                   'JT+NN')
        self.crud_object(body_nn_jt, self.url_ngt)

    def test_crud_ngt_tt_dn(self):
        """This test checks node group template creation with
        configuration | TT + DN |.
        """
        body_tt_dn = self.make_node_group_template('worker-dn-tt', 'qa probe',
                                                   'TT+DN')
        self.crud_object(body_tt_dn, self.url_ngt)

    def test_crud_ngt_nnttdn(self):
        """This test checks node group template creation with
        configuration | NN + TT + DN |.
        """
        body_nn_tt_dn = self.make_node_group_template('nn-dn-tt', 'qa probe',
                                                      'NN+TT+DN')
        self.crud_object(body_nn_tt_dn, self.url_ngt)

    def test_crud_ngt_jtttdn(self):
        """This test checks node group template creation with
        configuration | JT + TT + DN |.
        """
        body_jt_tt_dn = self.make_node_group_template('jt-dn-tt', 'qa probe',
                                                      'JT+TT+DN')
        self.crud_object(body_jt_tt_dn, self.url_ngt)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import os
import tempfile
import unittest2

from savanna import context
from savanna.db import api as db_api
from savanna.openstack.common.db.sqlalchemy import session
from savanna.openstack.common import timeutils
from savanna.openstack.common import uuidutils


class DbTestCase(unittest2.TestCase):

    def set_tenant(self, tenant_id="tenant_1"):
        context.set_ctx(
            context.Context('test_user', tenant_id, 'test_auth_token', {}))

    def setUp(self):
        self.set_tenant()
        self.db_fd, self.db_path = tempfile.mkstemp()
        session.set_defaults('sqlite:///' + self.db_path, self.db_path)
        db_api.configure_db()

    def tearDown(self):
        db_api.clear_db()
        os.close(self.db_fd)
        os.unlink(self.db_path)
        context.set_ctx(None)

    def assertIsValidModelObject(self, res):
        self.assertIsNotNone(res)
        self.assertIsNotNone(res.dict)
        self.assertTrue(uuidutils.is_uuid_like(res.id))

        # check created/updated
        delta = datetime.timedelta(seconds=2)
        now = timeutils.utcnow()

        self.assertAlmostEqual(res.created, now, delta=delta)
        self.assertAlmostEqual(res.updated, now, delta=delta)

    def get_clean_dict(self, res):
        res_dict = res.dict
        del res_dict['created']
        del res_dict['updated']
        del res_dict['id']
        if 'tenant_id' in res_dict:
            del res_dict['tenant_id']

        return res_dict



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import unittest2

import savanna.db.models as m
from savanna import exceptions as ex
from savanna.plugins.vanilla import plugin
from savanna.service import api
import savanna.service.validation as v
from savanna.service.validations import clusters_scaling as c_s
from savanna.tests.unit.service.validation import utils as u


def _get_plugin(plugin_name):
    if plugin_name == 'vanilla':
        return plugin.VanillaProvider()
    return None


class TestScalingValidation(unittest2.TestCase):
    def setUp(self):
        api.plugin_base.setup_plugins()
        self._create_object_fun = mock.Mock()

    @mock.patch('savanna.service.api.get_cluster')
    @mock.patch('savanna.plugins.base.PluginManager.get_plugin')
    def _assert_check_scaling(self,
                              get_plugin_p=None,
                              get_cluster_p=None,
                              data=None, cluster=None,
                              expected_message=None):

        get_cluster_p.return_value = cluster
        get_plugin_p.side_effect = _get_plugin

        with self.assertRaises(ex.InvalidException):
            try:
                c_s.check_cluster_scaling(data, cluster.id)
            except ex.InvalidException as e:
                self.assertEqual(expected_message, e.message)
                raise e

    def test_check_cluster_scaling_resize_ng(self):
        cluster = m.Cluster('test-cluster', 'tenant', 'vanilla', '1.2.2')
        cluster.status = 'Validating'
        ng = m.NodeGroup('ng', '42', ['namenode'], 1)
        cluster.node_groups.append(ng)
        self._assert_check_scaling(data={}, cluster=cluster,
                                   expected_message=
                                   "Cluster cannot be scaled "
                                   "not in 'Active' "
                                   "status. Cluster status: "
                                   "Validating")
        cluster.status = 'Active'
        data = {
            'resize_node_groups': [
                {
                    'name': 'a',
                    'flavor_id': '42',
                    'node_processes': ['namenode']
                }
            ],
        }
        self._assert_check_scaling(data=data, cluster=cluster,
                                   expected_message=
                                   "Cluster doesn't contain "
                                   "node group with name 'a'")
        data.update({'resize_node_groups': [
            {
                'name': 'a',
                'flavor_id': '42',
                'node_processes': ['namenode']
            },
            {
                'name': 'a',
                'flavor_id': '42',
                'node_processes': ['namenode']
            }
        ]})
        self._assert_check_scaling(data=data, cluster=cluster,
                                   expected_message=
                                   'Duplicates in node '
                                   'group names are detected')

    def test_check_cluster_scaling_add_ng(self):
        cluster = m.Cluster('test-cluster', 'tenant', 'vanilla', '1.2.2')
        ng = m.NodeGroup('ng', '42', ['namenode'], 1)
        cluster.node_groups.append(ng)
        cluster.status = 'Active'
        data = {
            'add_node_groups': [
                {
                    'name': 'a',
                    'flavor_id': '42',
                    'node_processes': ['namenode']
                },
                {
                    'name': 'a',
                    'flavor_id': '42',
                    'node_processes': ['namenode']
                }
            ]
        }
        self._assert_check_scaling(data=data, cluster=cluster,
                                   expected_message=
                                   'Duplicates in node '
                                   'group names are detected')
        data = {
            'add_node_groups': [
                {
                    'name': 'ng',
                    'flavor_id': '42',
                    'node_processes': ['namenode']
                },
            ]
        }
        self._assert_check_scaling(data=data, cluster=cluster,
                                   expected_message=
                                   "Can't add new nodegroup. "
                                   "Cluster already has nodegroup "
                                   "with name 'ng'")

    def _assert_calls(self, mock, call_info):
        if not call_info:
            self.assertEqual(mock.call_count, 0)
        else:
            self.assertEqual(mock.call_count, call_info[0])
            self.assertEqual(mock.call_args[0][0].code, call_info[1])
            self.assertEqual(mock.call_args[0][0].message, call_info[2])

    @mock.patch("savanna.utils.api.request_data")
    @mock.patch("savanna.utils.api.bad_request")
    def _assert_cluster_scaling_validation(self,
                                           bad_req=None,
                                           req_data=None,
                                           data=None,
                                           bad_req_i=None):
        m_func = mock.Mock()
        m_func.__name__ = "m_func"
        req_data.return_value = data
        v.validate(c_s.CLUSTER_SCALING_SCHEMA,
                   self._create_object_fun)(m_func)(data=data,
                                                    cluster_id='42')

        self.assertEqual(req_data.call_count, 1)
        self._assert_calls(bad_req, bad_req_i)

    def test_cluster_scaling_scheme_v_resize_ng(self):
        self._create_object_fun = mock.Mock()
        data = {
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u'{} is not valid under any of the given schemas')
        )
        data = {
            'resize_node_groups': [{}]
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"'name' is a required property")
        )
        data = {
            'resize_node_groups': [
                {
                    'name': 'a'
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"'count' is a required property")
        )

    def test_cluster_scaling_validation_add_ng(self):
        data = {
            'add_node_groups': [
                {
                    "node_group_template_id": "5185a809-6bf7-"
                                              "44ed-9de3-618270550e2c",
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       "{'node_group_template_id': "
                       "'5185a809-6bf7-44ed-9de3-618270550e2c'} "
                       "is not valid under any of the given schemas")
        )
        data = {
            'add_node_groups': [
                {
                    "node_group_template_id": "5185a809-6bf7-"
                                              "44ed-9de3-618270550e2c",
                    'name': 'a'
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"{'node_group_template_id': "
                       u"'5185a809-6bf7-44ed-9de3-618270550e2c', "
                       u"'name': 'a'} is not valid under any "
                       u"of the given schemas")
        )

    def test_cluster_scaling_validation_right_schema(self):
        data = {
            'add_node_groups': [
                {
                    "node_group_template_id": "5185a809-6bf7-"
                                              "44ed-9de3-618270550e2c",
                    'name': 'a',
                    'count': 3
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data
        )
        data = {
            'resize_node_groups': [
                {
                    'name': 'a',
                    'count': 3
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data
        )
        data = {
            'resize_node_groups': [
                {
                    'name': 'a',
                    'count': 3
                }
            ],
            'add_node_groups': [
                {
                    "node_group_template_id": "5185a809-6bf7-"
                                              "44ed-9de3-618270550e2c",
                    'name': 'a',
                    'count': 3
                }
            ]
        }
        self._assert_cluster_scaling_validation(
            data=data
        )

    def test_cluster_scaling_scheme_validation_types(self):
        data = {
            'resize_node_groups': {},
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"{} is not of type 'array'")
        )
        data = {
            'add_node_groups': {}
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u"{} is not of type 'array'")
        )
        data = {
            'resize_node_groups': [],
        }
        self._assert_cluster_scaling_validation(
            data=data,
            bad_req_i=(1, 'VALIDATION_ERROR',
                       u'[] is too short')
        )

    def test_cluster_scaling_v_right_data(self):
        self._create_object_fun = c_s.check_cluster_scaling
        cluster = m.Cluster('test-cluster', 'tenant', 'vanilla', '1.2.2')
        ng = m.NodeGroup('ng', '42', ['namenode'], 1)
        cluster.node_groups.append(ng)
        cluster.status = 'Active'

        data = {
            'resize_node_groups': [
                {
                    'name': 'ng',
                    'count': 4
                }
            ],
            'add_node_groups': [
                {
                    'name': 'a',
                    'flavor_id': '42',
                    'node_processes': ['namenode'],
                    'count': 3
                },
            ]
        }
        patchers = u.start_patch()
        self._assert_cluster_scaling_validation(data=data)
        u.stop_patch(patchers)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
import flask
from keystoneclient.middleware import auth_token
from oslo.config import cfg
from werkzeug import exceptions as werkzeug_exceptions

from savanna.api import v10 as api_v10
from savanna import context
from savanna.db import api as db_api
from savanna.middleware import auth_valid
from savanna.openstack.common import log
from savanna.openstack.common.middleware import debug
from savanna.plugins import base as plugins_base
from savanna.utils import api as api_utils
from savanna.utils import patches
from savanna.utils import scheduler

LOG = log.getLogger(__name__)

eventlet.monkey_patch(
    os=True, select=True, socket=True, thread=True, time=True)

# Patches minidom's writexml to avoid excess whitespaces in generated xml
# configuration files that brakes Hadoop.
patches.patch_minidom_writexml()

opts = [
    cfg.StrOpt('os_auth_protocol',
               default='http',
               help='Protocol used to access OpenStack Identity service'),
    cfg.StrOpt('os_auth_host',
               default='127.0.0.1',
               help='IP or hostname of machine on which OpenStack Identity '
                    'service is located'),
    cfg.StrOpt('os_auth_port',
               default='35357',
               help='Port of OpenStack Identity service'),
    cfg.StrOpt('os_admin_username',
               default='admin',
               help='This OpenStack user is used to verify provided tokens. '
                    'The user must have admin role in <os_admin_tenant_name> '
                    'tenant'),
    cfg.StrOpt('os_admin_password',
               default='nova',
               help='Password of the admin user'),
    cfg.StrOpt('os_admin_tenant_name',
               default='admin',
               help='Name of tenant where the user is admin')
]

CONF = cfg.CONF
CONF.register_opts(opts)


def make_app():
    """App builder (wsgi)

    Entry point for Savanna REST API server
    """
    app = flask.Flask('savanna.api')

    @app.route('/', methods=['GET'])
    def version_list():
        context.set_ctx(None)
        return api_utils.render({
            "versions": [
                {"id": "v1.0", "status": "CURRENT"}
            ]
        })

    @app.teardown_request
    def teardown_request(_ex=None):
        # TODO(slukjanov): how it'll work in case of exception?
        if context.has_ctx():
            if flask.request.path != '/':
                try:
                    session = context.session()
                    if session.transaction:
                        session.transaction.commit()
                except Exception as e:
                    return api_utils.internal_error(
                        500, 'Internal Server Error', e)

        context.set_ctx(None)

    app.register_blueprint(api_v10.rest, url_prefix='/v1.0')

    db_api.configure_db()
    scheduler.setup_scheduler(app)
    plugins_base.setup_plugins()

    def make_json_error(ex):
        status_code = (ex.code
                       if isinstance(ex, werkzeug_exceptions.HTTPException)
                       else 500)
        description = (ex.description
                       if isinstance(ex, werkzeug_exceptions.HTTPException)
                       else str(ex))
        return api_utils.render({'error': status_code,
                                 'error_message': description},
                                status=status_code)

    for code in werkzeug_exceptions.default_exceptions.iterkeys():
        app.error_handler_spec[None][code] = make_json_error

    if CONF.debug and not CONF.log_exchange:
        LOG.debug('Logging of request/response exchange could be enabled using'
                  ' flag --log-exchange')

    if CONF.log_exchange:
        app.wsgi_app = debug.Debug.factory(app.config)(app.wsgi_app)

    app.wsgi_app = auth_valid.filter_factory(app.config)(app.wsgi_app)

    app.wsgi_app = auth_token.filter_factory(
        app.config,
        auth_host=CONF.os_auth_host,
        auth_port=CONF.os_auth_port,
        auth_protocol=CONF.os_auth_protocol,
        admin_user=CONF.os_admin_username,
        admin_password=CONF.os_admin_password,
        admin_tenant_name=CONF.os_admin_tenant_name
    )(app.wsgi_app)

    return app



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from savanna import exceptions as ex
from savanna.service import api
import savanna.service.validations.base as b
import savanna.service.validations.node_group_templates as ng_tml


def _build_ng_schema_for_cluster_tmpl():
    cl_tmpl_ng_schema = copy.deepcopy(ng_tml.NODE_GROUP_TEMPLATE_SCHEMA)
    cl_tmpl_ng_schema['properties'].update({"count": {"type": "integer"}})
    cl_tmpl_ng_schema["required"] = ['name', 'flavor_id',
                                     'node_processes', 'count']
    del cl_tmpl_ng_schema['properties']['hadoop_version']
    del cl_tmpl_ng_schema['properties']['plugin_name']
    return cl_tmpl_ng_schema


_cluster_tmpl_ng_schema = _build_ng_schema_for_cluster_tmpl()


def _build_ng_tmpl_schema_for_cluster_template():
    cl_tmpl_ng_tmpl_schema = copy.deepcopy(_cluster_tmpl_ng_schema)
    cl_tmpl_ng_tmpl_schema['properties'].update(
        {
            "node_group_template_id": {
                "type": "string",
                "format": "uuid",
            }
        })
    cl_tmpl_ng_tmpl_schema["required"] = ["node_group_template_id",
                                          "name", "count"]
    return cl_tmpl_ng_tmpl_schema


_cluster_tmpl_ng_tmpl_schema = _build_ng_tmpl_schema_for_cluster_template()

CLUSTER_TEMPLATE_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
            "minLength": 1,
            "maxLength": 50,
            "format": "valid_name",
        },
        "plugin_name": {
            "type": "string",
        },
        "hadoop_version": {
            "type": "string",
        },
        "default_image_id": {
            "type": "string",
            "format": "uuid",
        },
        "cluster_configs": {
            "type": "configs",
        },
        "node_groups": {
            "type": "array",
            "items": {
                "oneOf": [_cluster_tmpl_ng_tmpl_schema,
                          _cluster_tmpl_ng_schema]
            }
        },
        "anti_affinity": {
            "type": "array",
            "items": {
                "type": "string",
            },
        },
        "description": {
            "type": "string",
        },
    },
    "additionalProperties": False,
    "required": [
        "name",
        "plugin_name",
        "hadoop_version",
    ]
}


def check_cluster_template_create(data, **kwargs):
    b.check_cluster_template_unique_name(data['name'])
    b.check_plugin_name_exists(data['plugin_name'])
    b.check_plugin_supports_version(data['plugin_name'],
                                    data['hadoop_version'])

    if data.get('default_image_id'):
        b.check_image_registered(data['default_image_id'])
        b.check_required_image_tags(data['plugin_name'],
                                    data['hadoop_version'],
                                    data['default_image_id'])

    b.check_all_configurations(data)

    if data.get('anti_affinity'):
        b.check_node_processes(data['plugin_name'], data['hadoop_version'],
                               data['anti_affinity'])


def check_cluster_template_usage(cluster_template_id, **kwargs):
    clusters = api.get_clusters()
    use_cluster_template_ids = [cluster.cluster_template_id
                                for cluster in clusters]

    if cluster_template_id in use_cluster_template_ids:
        raise ex.InvalidException(
            "Cluster template %s is use" % cluster_template_id)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

image_register_schema = {
    "type": "object",
    "properties": {
        "username": {
            "type": "string",
        },
        "description": {
            "type": "string",
        },
    },
    "additionalProperties": False,
    "required": ["username"]
}

image_tags_schema = {
    "type": "object",
    "properties": {
        "tags": {
            "type": "array",
            "items": {
                "type": "string",
                "format": "valid_tag"
            },
        },
    },
    "additionalProperties": False,
    "required": ["tags"]
}


def check_image_register(data, **kwargs):
    pass


def check_tags(data, **kwargs):
    pass



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from novaclient import exceptions as nova_exceptions

from savanna import context
from savanna.db import models as m
from savanna.openstack.common import excutils
from savanna.openstack.common import log as logging
from savanna.service import networks
from savanna.service import volumes
from savanna.utils import crypto
from savanna.utils.openstack import nova

LOG = logging.getLogger(__name__)


def create_cluster(cluster):
    try:
        # create all instances
        context.model_update(cluster, status='Spawning')
        _create_instances(cluster)

        # wait for all instances are up and accessible
        context.model_update(cluster, status='Waiting')
        _await_instances(cluster)

        # attach volumes
        volumes.attach(cluster)

        # prepare all instances
        context.model_update(cluster, status='Preparing')
        _configure_instances(cluster)
    except Exception as ex:
        LOG.warn("Can't start cluster '%s' (reason: %s)", cluster.name, ex)
        with excutils.save_and_reraise_exception():
            context.model_update(cluster, status='Error',
                                 status_description=str(ex))
            _rollback_cluster_creation(cluster, ex)


def scale_cluster(cluster, node_group_names_map, plugin):
    # Now let's work with real node_groups, not names:
    node_groups_map = {}
    for ng in cluster.node_groups:
        if ng.name in node_group_names_map:
            node_groups_map.update({ng: node_group_names_map[ng.name]})
    instances_list = []
    try:
        instances_list = _scale_cluster_instances(
            cluster, node_groups_map, plugin)
        _clean_cluster_from_empty_ng(cluster)
        _await_instances(cluster)
        volumes.attach_to_instances(instances_list)

    except Exception as ex:
        LOG.warn("Can't scale cluster '%s' (reason: %s)", cluster.name, ex)
        with excutils.save_and_reraise_exception():
            _rollback_cluster_scaling(cluster, instances_list, ex)
            instances_list = []
            _clean_cluster_from_empty_ng(cluster)
            if cluster.status == 'Decommissioning':
                context.model_update(cluster, status='Error')
            else:
                context.model_update(cluster, status='Active')
    # we should be here with valid cluster: if instances creation
    # was not successful all extra-instances will be removed above
    if instances_list:
        _configure_instances(cluster)
    return instances_list


def _generate_anti_affinity_groups(cluster):
    aa_groups = {}

    for node_group in cluster.node_groups:
        for instance in node_group.instances:
            if instance.instance_id:
                for process in node_group.node_processes:
                    if process in cluster.anti_affinity:
                        aa_group = aa_groups.get(process, [])
                        aa_group.append(instance.instance_id)
                        aa_groups[process] = aa_group

    return aa_groups


def _create_instances(cluster):
    aa_groups = _generate_anti_affinity_groups(cluster)
    for node_group in cluster.node_groups:
        userdata = _generate_user_data_script(node_group)
        for idx in xrange(1, node_group.count + 1):
            _run_instance(cluster, node_group, idx, aa_groups, userdata)


def _scale_cluster_instances(cluster, node_groups_map, plugin):
    aa_groups = _generate_anti_affinity_groups(cluster)
    instances_to_delete = []
    node_groups_to_enlarge = []

    for node_group in node_groups_map:
        count = node_groups_map[node_group]
        if count < node_group.count:
            instances_to_delete += node_group.instances[count:node_group.count]
        else:
            node_groups_to_enlarge.append(node_group)

    if instances_to_delete:
        cluster.status = 'Decommissioning'
        plugin.decommission_nodes(cluster, instances_to_delete)
        cluster.status = 'Deleting Instances'
        for instance in instances_to_delete:
            node_group = instance.node_group
            node_group.instances.remove(instance)
            _shutdown_instance(instance)
            node_group.count -= 1
            context.model_save(node_group)

    instances_to_add = []
    if node_groups_to_enlarge:
        cluster.status = 'Adding Instances'
        for node_group in node_groups_to_enlarge:
            count = node_groups_map[node_group]
            userdata = _generate_user_data_script(node_group)
            for idx in xrange(node_group.count + 1, count + 1):
                instance = _run_instance(cluster, node_group, idx,
                                         aa_groups, userdata)
                instances_to_add.append(instance)
            node_group.count = count

    return instances_to_add


def _run_instance(cluster, node_group, idx, aa_groups, userdata):
    """Create instance using nova client and persist them into DB."""
    session = context.ctx().session
    name = '%s-%s-%03d' % (cluster.name, node_group.name, idx)

    # aa_groups: node process -> instance ids
    aa_ids = []
    for node_process in node_group.node_processes:
        aa_ids += aa_groups.get(node_process) or []

    # create instances only at hosts w/ no instances w/ aa-enabled processes
    hints = {'different_host': list(set(aa_ids))} if aa_ids else None

    context.model_save(node_group)

    nova_instance = nova.client().servers.create(
        name, node_group.get_image_id(), node_group.flavor_id,
        scheduler_hints=hints, userdata=userdata,
        key_name=cluster.user_keypair_id)

    with session.begin():
        instance = m.Instance(node_group.id, nova_instance.id, name)
        node_group.instances.append(instance)
        session.add(instance)

    # save instance id to aa_groups to support aa feature
    for node_process in node_group.node_processes:
        if node_process in cluster.anti_affinity:
            aa_group_ids = aa_groups.get(node_process, [])
            aa_group_ids.append(nova_instance.id)
            aa_groups[node_process] = aa_group_ids

    return instance


def _generate_user_data_script(node_group):
    script_template = """#!/bin/bash
echo "%(public_key)s" >> %(user_home)s/.ssh/authorized_keys
echo "%(private_key)s" > %(user_home)s/.ssh/id_rsa
"""
    cluster = node_group.cluster
    if node_group.username == "root":
        user_home = "/root/"
    else:
        user_home = "/home/%s/" % node_group.username

    return script_template % {
        "public_key": crypto.private_key_to_public_key(cluster.private_key),
        "private_key": cluster.private_key,
        "user_home": user_home
    }


def _await_instances(cluster):
    """Await all instances are in Active status and available."""
    all_up = False
    while not all_up:
        all_up = True
        for node_group in cluster.node_groups:
            for instance in node_group.instances:
                if not _check_if_up(instance):
                    all_up = False
        context.sleep(1)


def _check_if_up(instance):
    if hasattr(instance, '_is_up'):
        return True

    server = instance.nova_info
    if server.status == 'ERROR':
        # TODO(slukjanov): replace with specific error
        raise RuntimeError("node %s has error status" % server.name)

    if server.status != 'ACTIVE':
        return False

    if len(server.networks) == 0:
        return False

    if not networks.init_instances_ips(instance, server):
        return False

    try:
        exit_code, _ = instance.remote.execute_command("hostname")
        if exit_code:
            return False
    except Exception as ex:
        LOG.debug("Can't login to node %s (%s), reason %s",
                  server.name, instance.management_ip, ex)
        return False

    instance._is_up = True
    return True


def _configure_instances(cluster):
    """Configure active instances.

    * generate /etc/hosts
    * setup passwordless login
    * etc.
    """
    hosts = _generate_etc_hosts(cluster)
    for node_group in cluster.node_groups:
        for instance in node_group.instances:
            with instance.remote as remote:
                remote.write_file_to('etc-hosts', hosts)
                remote.execute_command('sudo mv etc-hosts /etc/hosts')
                remote.execute_command('chmod 400 .ssh/id_rsa')


def _generate_etc_hosts(cluster):
    hosts = "127.0.0.1 localhost\n"
    for node_group in cluster.node_groups:
        for instance in node_group.instances:
            hosts += "%s %s %s\n" % (instance.internal_ip,
                                     instance.fqdn,
                                     instance.hostname)

    return hosts


def _rollback_cluster_creation(cluster, ex):
    """Shutdown all instances and update cluster status."""
    LOG.info("Cluster '%s' creation rollback (reason: %s)", cluster.name, ex)

    session = context.ctx().session
    _shutdown_instances(cluster, True)
    volumes.detach(cluster)
    alive_instances = set([srv.id for srv in nova.client().servers.list()])

    for node_group in cluster.node_groups:
        for instance in node_group.instances:
            if instance.instance_id in alive_instances:
                nova.client().servers.delete(instance.instance_id)
            with session.begin():
                session.delete(instance)


def _rollback_cluster_scaling(cluster, instances, ex):
    """Attempt to rollback cluster scaling."""
    LOG.info("Cluster '%s' scaling rollback (reason: %s)", cluster.name, ex)
    try:
        volumes.detach_from_instances(instances)
    except Exception:
        raise
    finally:
        for i in instances:
            ng = i.node_group
            _shutdown_instance(i)
            ng.count -= 1


def _shutdown_instances(cluster, quiet=False):
    for node_group in cluster.node_groups:
        for instance in node_group.instances:
            _shutdown_instance(instance)


def _shutdown_instance(instance):
    session = context.ctx().session
    try:
        nova.client().servers.delete(instance.instance_id)
    except nova_exceptions.NotFound:
        #Just ignore non-existing instances
        pass

    with session.begin():
        session.delete(instance)


def shutdown_cluster(cluster):
    """Shutdown specified cluster and all related resources."""
    volumes.detach(cluster)
    _shutdown_instances(cluster)


def _clean_cluster_from_empty_ng(cluster):
    session = context.ctx().session
    with session.begin():
        all_ng = cluster.node_groups
    for ng in all_ng:
        if ng.count == 0:
            session.delete(ng)
            cluster.node_groups.remove(ng)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2010-2011 OpenStack Foundation.
# Copyright 2012 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Implementation of paginate query."""

import sqlalchemy

from savanna.openstack.common.gettextutils import _
from savanna.openstack.common import log as logging


LOG = logging.getLogger(__name__)


class InvalidSortKey(Exception):
    message = _("Sort key supplied was not valid.")


# copy from glance/db/sqlalchemy/api.py
def paginate_query(query, model, limit, sort_keys, marker=None,
                   sort_dir=None, sort_dirs=None):
    """Returns a query with sorting / pagination criteria added.

    Pagination works by requiring a unique sort_key, specified by sort_keys.
    (If sort_keys is not unique, then we risk looping through values.)
    We use the last row in the previous page as the 'marker' for pagination.
    So we must return values that follow the passed marker in the order.
    With a single-valued sort_key, this would be easy: sort_key > X.
    With a compound-values sort_key, (k1, k2, k3) we must do this to repeat
    the lexicographical ordering:
    (k1 > X1) or (k1 == X1 && k2 > X2) or (k1 == X1 && k2 == X2 && k3 > X3)

    We also have to cope with different sort_directions.

    Typically, the id of the last row is used as the client-facing pagination
    marker, then the actual marker object must be fetched from the db and
    passed in to us as marker.

    :param query: the query object to which we should add paging/sorting
    :param model: the ORM model class
    :param limit: maximum number of items to return
    :param sort_keys: array of attributes by which results should be sorted
    :param marker: the last item of the previous page; we returns the next
                    results after this value.
    :param sort_dir: direction in which results should be sorted (asc, desc)
    :param sort_dirs: per-column array of sort_dirs, corresponding to sort_keys

    :rtype: sqlalchemy.orm.query.Query
    :return: The query with sorting/pagination added.
    """

    if 'id' not in sort_keys:
        # TODO(justinsb): If this ever gives a false-positive, check
        # the actual primary key, rather than assuming its id
        LOG.warn(_('Id not in sort_keys; is sort_keys unique?'))

    assert(not (sort_dir and sort_dirs))

    # Default the sort direction to ascending
    if sort_dirs is None and sort_dir is None:
        sort_dir = 'asc'

    # Ensure a per-column sort direction
    if sort_dirs is None:
        sort_dirs = [sort_dir for _sort_key in sort_keys]

    assert(len(sort_dirs) == len(sort_keys))

    # Add sorting
    for current_sort_key, current_sort_dir in zip(sort_keys, sort_dirs):
        sort_dir_func = {
            'asc': sqlalchemy.asc,
            'desc': sqlalchemy.desc,
        }[current_sort_dir]

        try:
            sort_key_attr = getattr(model, current_sort_key)
        except AttributeError:
            raise InvalidSortKey()
        query = query.order_by(sort_dir_func(sort_key_attr))

    # Add pagination
    if marker is not None:
        marker_values = []
        for sort_key in sort_keys:
            v = getattr(marker, sort_key)
            marker_values.append(v)

        # Build up an array of sort criteria as in the docstring
        criteria_list = []
        for i in range(0, len(sort_keys)):
            crit_attrs = []
            for j in range(0, i):
                model_attr = getattr(model, sort_keys[j])
                crit_attrs.append((model_attr == marker_values[j]))

            model_attr = getattr(model, sort_keys[i])
            if sort_dirs[i] == 'desc':
                crit_attrs.append((model_attr < marker_values[i]))
            elif sort_dirs[i] == 'asc':
                crit_attrs.append((model_attr > marker_values[i]))
            else:
                raise ValueError(_("Unknown sort direction, "
                                   "must be 'desc' or 'asc'"))

            criteria = sqlalchemy.sql.and_(*crit_attrs)
            criteria_list.append(criteria)

        f = sqlalchemy.sql.or_(*criteria_list)
        query = query.filter(f)

    if limit is not None:
        query = query.limit(limit)

    return query



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna import context as ctx
from savanna.db import models as m
from savanna.tests.unit import base as models_test_base


class ClusterModelTest(models_test_base.DbTestCase):
    def testCreateCluster(self):
        session = ctx.current().session
        with session.begin():
            c = m.Cluster('c-1', 't-1', 'p-1', 'hv-1')
            session.add(c)

        with session.begin():
            res = session.query(m.Cluster).filter_by().first()

            self.assertIsValidModelObject(res)

    def testCreateClusterFromDict(self):
        c = m.Cluster('c-1', 't-1', 'p-1', 'hv-1')
        c_dict = c.dict
        del c_dict['created']
        del c_dict['updated']
        del c_dict['id']
        del c_dict['node_groups']
        del c_dict['status']
        del c_dict['status_description']
        del c_dict['info']

        c_dict.update({
            'tenant_id': 't-1'
        })
        self.assertEqual(self.get_clean_dict(c),
                         self.get_clean_dict(m.Cluster(**c_dict)))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import telnetlib

from savanna.tests.integration import base
import savanna.tests.integration.configs.parameters as param


def _add_config(body, config):
    if config in [param.NAMENODE_CONFIG, param.DATANODE_CONFIG]:
        body['node_configs']['HDFS'] = config
    elif config == param.GENERAL_CONFIG:
        body['cluster_configs']['general'] = config
    elif config == param.CLUSTER_HDFS_CONFIG:
        body['cluster_configs']['HDFS'] = config
    elif config == param.CLUSTER_MAPREDUCE_CONFIG:
        body['cluster_configs']['MapReduce'] = config
    else:
        body['node_configs']['MapReduce'] = config


class ClusterConfigTest(base.ITestCase):

    def setUp(self):
        super(ClusterConfigTest, self).setUp()
        telnetlib.Telnet(self.host, self.port)

    def assertConfigs(self, get_config, param_config):
        self.assertEqual(get_config, param_config,
                         msg='configs are not equal: %s != %s'
                             % (str(get_config), str(param_config)))

    def assertConfigOnNode(self, host, config, value):
        conf = config.replace(' ', '')
        com = self.execute_command(host, './script.sh %s -val %s -url %s' %
                                         (conf, value, param.OS_AUTH_URL))
        self.assertEqual(com[0], 0,
                         msg='host: %s, config %s is not equal: %s'
                             % (host, config, value))

    def _cluster_config_testing(self, cluster_body):
        cluster_id = None
        try:
            _add_config(cluster_body, param.GENERAL_CONFIG)
            _add_config(cluster_body, param.CLUSTER_HDFS_CONFIG)
            _add_config(cluster_body, param.CLUSTER_MAPREDUCE_CONFIG)
            cluster_id = self.create_cluster_and_get_id(cluster_body)
            get_data = self.get_object(self.url_cluster_with_slash,
                                       cluster_id, 200, True)
            get_data = get_data['cluster']
            self.assertConfigs(get_data['cluster_configs']['general'],
                               param.GENERAL_CONFIG)
            self.assertConfigs(get_data['cluster_configs']['HDFS'],
                               param.CLUSTER_HDFS_CONFIG)
            self.assertConfigs(get_data['cluster_configs']['MapReduce'],
                               param.CLUSTER_MAPREDUCE_CONFIG)
            node_groups = get_data['node_groups']
            ip_instances = {}
            process_map = {
                'namenode': {
                    'service': 'HDFS', 'param': param.NAMENODE_CONFIG},
                'jobtracker': {
                    'service': 'MapReduce', 'param': param.JOBTRACKER_CONFIG},
                'datanode': {
                    'service': 'HDFS', 'param': param.DATANODE_CONFIG},
                'tasktracker': {
                    'service': 'MapReduce', 'param': param.TASKTRACKER_CONFIG}
            }

            def get_node_configs(node_group, process):
                return \
                    node_group['node_configs'][process_map[process]['service']]

            def get_param(process):
                return process_map[process]['param']

            for node_group in node_groups:
                for process in node_group['node_processes']:
                    self.assertConfigs(
                        get_node_configs(node_group,
                                         process), get_param(process))
                instances = node_group['instances']
                for instans in instances:
                    management_ip = instans['management_ip']
                    self.transfer_script_to_node(
                        management_ip, 'test_config/config_test_script.sh')
                    ip_instances[management_ip] = node_group[
                        'node_processes']
            try:
                for key, processes in ip_instances.items():
                    telnetlib.Telnet(key, '22')
                    for conf, value in param.CLUSTER_MAPREDUCE_CONFIG.items():
                        self.assertConfigOnNode(key, conf, value)
                    for conf, value in param.CLUSTER_HDFS_CONFIG.items():
                        self.assertConfigOnNode(key, conf, value)
                    for process in processes:
                        for sec_key, sec_value in get_param(process).items():
                            self.assertConfigOnNode(key, sec_key, sec_value)
                    if 'namenode' in processes:
                        for sec_key, sec_value in param.GENERAL_CONFIG.items():
                            self.assertConfigOnNode(
                                key, sec_key, sec_value)
            except Exception as e:
                self.fail(e.message)
        except Exception as e:
            self.fail(e.message)
        finally:
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)

    def test_cluster_config_nnjt_ttdn(self):
        id_master_ngt = None
        id_worker_ngt = None
        try:
            master_ngt_body = self.make_node_group_template(
                'master-ngt', 'qa probe', 'JT+NN')
            _add_config(master_ngt_body, param.NAMENODE_CONFIG)
            _add_config(master_ngt_body, param.JOBTRACKER_CONFIG)
            id_master_ngt = self.get_object_id(
                'node_group_template', self.post_object(self.url_ngt,
                                                        master_ngt_body, 202))

            worker_ngt_body = self.make_node_group_template(
                'worker-ngt', 'qa probe', 'TT+DN')
            _add_config(worker_ngt_body, param.DATANODE_CONFIG)
            _add_config(worker_ngt_body, param.TASKTRACKER_CONFIG)
            id_worker_ngt = self.get_object_id(
                'node_group_template', self.post_object(self.url_ngt,
                                                        worker_ngt_body, 202))

            ngt_id_list = {id_master_ngt: 1, id_worker_ngt: 2}
            cl_body = self.make_cl_body_node_group_templates(ngt_id_list)
            self._cluster_config_testing(cl_body)
        except Exception as e:
            self.fail(str(e))
        finally:
            self.del_object(self.url_ngt_with_slash, id_master_ngt, 204)
            self.del_object(self.url_ngt_with_slash, id_worker_ngt, 204)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlalchemy as sql

from savanna.db import model_base
from savanna.openstack.common.db.sqlalchemy import session
from savanna.openstack.common import log as logging

LOG = logging.getLogger(__name__)

_DB_ENGINE = None


def configure_db():
    """Configure database.

    Establish the database, create an engine if needed, and register
    the models.
    """
    global _DB_ENGINE
    if not _DB_ENGINE:
        session.set_defaults('sqlite:////tmp/savanna-server.db',
                             'savanna-server.db')
        _DB_ENGINE = session.get_engine(sqlite_fk=True)
        register_models()


def clear_db(base=model_base.SavannaBase):
    global _DB_ENGINE
    unregister_models(base)
    session.cleanup()
    _DB_ENGINE = None


def get_session(autocommit=True, expire_on_commit=False):
    """Helper method to grab session."""
    return session.get_session(autocommit=autocommit,
                               expire_on_commit=expire_on_commit,
                               sqlite_fk=True)


def register_models(base=model_base.SavannaBase):
    """Register Models and create properties."""
    try:
        engine = session.get_engine(sqlite_fk=True)
        base.metadata.create_all(engine)
    except sql.exc.OperationalError as e:
        LOG.info("Database registration exception: %s", e)
        return False
    return True


def unregister_models(base=model_base.SavannaBase):
    """Unregister Models, useful clearing out data before testing."""
    try:
        engine = session.get_engine(sqlite_fk=True)
        base.metadata.drop_all(engine)
    except Exception:
        LOG.exception("Database exception")



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest2

from savanna.plugins import provisioning as p


class ProvisioningPluginBaseTest(unittest2.TestCase):
    def test__map_to_user_inputs_success(self):
        c1, c2, c3, plugin = _build_configs_and_plugin()

        user_inputs = plugin._map_to_user_inputs(None, {
            'at-1': {
                'n-1': 'v-1',
                'n-3': 'v-3',
            },
            'at-2': {
                'n-2': 'v-2',
            },
        })

        self.assertEqual(user_inputs, [
            p.UserInput(c3, 'v-3'),
            p.UserInput(c1, 'v-1'),
            p.UserInput(c2, 'v-2'),
        ])

    def test__map_to_user_inputs_failure(self):
        c1, c2, c3, plugin = _build_configs_and_plugin()

        with self.assertRaises(RuntimeError):
            plugin._map_to_user_inputs(None, {
                'at-X': {
                    'n-1': 'v-1',
                },
            })

        with self.assertRaises(RuntimeError):
            plugin._map_to_user_inputs(None, {
                'at-1': {
                    'n-X': 'v-1',
                },
            })


def _build_configs_and_plugin():
    c1 = p.Config('n-1', 'at-1', 'cluster')
    c2 = p.Config('n-2', 'at-2', 'cluster')
    c3 = p.Config('n-3', 'at-1', 'node')

    class TestPlugin(TestEmptyPlugin):
        def get_configs(self, hadoop_version):
            return [c1, c2, c3]

    return c1, c2, c3, TestPlugin()


class TestEmptyPlugin(p.ProvisioningPluginBase):
    def get_title(self):
        pass

    def get_versions(self):
        pass

    def get_configs(self, hadoop_version):
        pass

    def get_node_processes(self, hadoop_version):
        pass

    def configure_cluster(self, cluster):
        pass

    def start_cluster(self, cluster):
        pass



# Copyright (c) 2013 Hortonworks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from savanna.plugins.hdp import clusterspec as cs
import unittest2


class ClusterSpecTest(unittest2.TestCase):
    service_validators = {}

    #TODO(jspeidel): test host manifest

    def test_parse_default_with_hosts(self):
        with open(os.path.join(os.path.realpath('../plugins'), 'hdp',
                               'resources',
                               'default-cluster.template'), 'r') as f:
            cluster_config_file = f.read()

        servers = []
        server1 = TestServer('host1', 'master', '11111', 3, '111.11.1111',
                             '222.11.1111',
                             node_processes=["namenode", "jobtracker",
                                             "secondary_namenode",
                                             "ganglia_server",
                                             "ganglia_monitor",
                                             "nagios_server", "AMBARI_SERVER",
                                             "ambari_agent"])
        server2 = TestServer('host2', 'slave', '11111', 3, '222.22.2222',
                             '333.22.2222')
        servers.append(server1)
        servers.append(server2)

        cluster = TestCluster()
        cluster.instances = servers

        cluster_config = cs.ClusterSpec(cluster_config_file, cluster)

        self._assert_services(cluster_config.services)
        self._assert_configurations(cluster_config.configurations)
        self._assert_host_role_mappings(cluster_config.node_groups)

        return cluster_config

    def test_select_correct_server_for_ambari_host(self):
        with open(os.path.join(os.path.realpath('../plugins'), 'hdp',
                               'resources',
                               'default-cluster.template'), 'r') as f:
            cluster_config_file = f.read()

        servers = []
        server1 = TestServer('ambari_machine', 'master', '11111', 3,
                             '111.11.1111', '222.11.1111',
                             node_processes=["namenode", "jobtracker",
                                             "secondary_namenode",
                                             "ganglia_server",
                                             "ganglia_monitor",
                                             "nagios_server", "AMBARI_SERVER",
                                             "ambari_agent"])
        server2 = TestServer('host2', 'slave', '11111', 3, '222.22.2222',
                             '333.22.2222',
                             node_processes=["datanode", "tasktracker",
                                             "ganglia_monitor", "hdfs_client",
                                             "mapreduce_client",
                                             "ambari_agent"])
        servers.append(server1)
        servers.append(server2)

        cluster = TestCluster
        cluster.instances = servers

        cluster_config = cs.ClusterSpec(cluster_config_file, cluster)
        self.assertIn('ambari_machine', cluster_config.str,
                      'Ambari host not found')

    def test_parse_default(self):
        with open(os.path.join(os.path.realpath('../plugins'), 'hdp',
                               'resources',
                               'default-cluster.template'), 'r') as f:
            cluster_config_file = f.read()

        cluster_config = cs.ClusterSpec(cluster_config_file)

        self._assert_services(cluster_config.services)
        self._assert_configurations(cluster_config.configurations)
        self._assert_host_role_mappings(cluster_config.node_groups)

        return cluster_config

    def test_normalize(self):
        with open(os.path.join(os.path.realpath('../plugins'), 'hdp',
                               'resources',
                               'default-cluster.template'), 'r') as f:
            cluster_config_file = f.read()

        cluster_config = cs.ClusterSpec(cluster_config_file)
        cluster = cluster_config.normalize()

        configs = cluster.cluster_configs
        contains_dfs_datanode_http_address = False
        contains_mapred_jobtracker_taskScheduler = False
        contains_dfs_include = False

        for entry in configs:
            config = entry.config
            # assert some random configurations across targets
            if config.name == 'dfs.datanode.http.address':
                contains_dfs_datanode_http_address = True
                self.assertEquals('string', config.type)
                self.assertEquals('0.0.0.0:50075', config.default_value)
                self.assertEquals('service:HDFS', config.applicable_target)

            if config.name == 'mapred.jobtracker.taskScheduler':
                contains_mapred_jobtracker_taskScheduler = True
                self.assertEquals('string', config.type)
                self.assertEquals(
                    'org.apache.hadoop.mapred.CapacityTaskScheduler',
                    config.default_value)
                self.assertEquals('service:MAPREDUCE',
                                  config.applicable_target)

            if config.name == 'dfs_include':
                contains_dfs_include = True
                self.assertEquals('string', config.type)
                self.assertEquals('dfs.include', config.default_value)
                self.assertEquals('service:HDFS', config.applicable_target)

                #            print 'Config: name: {0}, type:{1},
                # default value:{2}, target:{3}, Value:{4}'.format(
                #                config.name, config.type,
                # config.default_value,
                #  config.applicable_target, entry.value)

        self.assertTrue(contains_dfs_datanode_http_address)
        self.assertTrue(contains_mapred_jobtracker_taskScheduler)
        self.assertTrue(contains_dfs_include)
        node_groups = cluster.node_groups
        self.assertEquals(2, len(node_groups))
        contains_master_group = False
        contains_slave_group = False
        for i in range(2):
            node_group = node_groups[i]
            components = node_group.node_processes
            if node_group.name == "MASTER":
                contains_master_group = True
                self.assertEquals(8, len(components))
                self.assertIn('NAMENODE', components)
                self.assertIn('JOBTRACKER', components)
                self.assertIn('SECONDARY_NAMENODE', components)
                self.assertIn('GANGLIA_SERVER', components)
                self.assertIn('GANGLIA_MONITOR', components)
                self.assertIn('NAGIOS_SERVER', components)
                self.assertIn('AMBARI_SERVER', components)
                self.assertIn('AMBARI_AGENT', components)
                #TODO(jspeidel): node configs
                #TODO(jspeidel): vm_requirements
            elif node_group.name == 'SLAVE':
                contains_slave_group = True
                self.assertEquals(6, len(components))
                self.assertIn('DATANODE', components)
                self.assertIn('TASKTRACKER', components)
                self.assertIn('GANGLIA_MONITOR', components)
                self.assertIn('HDFS_CLIENT', components)
                self.assertIn('MAPREDUCE_CLIENT', components)
                self.assertIn('AMBARI_AGENT', components)
                #TODO(jspeidel): node configs
                #TODO(jspeidel): vm requirements
            else:
                self.fail('Unexpected node group: {0}'.format(node_group.name))
        self.assertTrue(contains_master_group)
        self.assertTrue(contains_slave_group)

    def _assert_services(self, services):
        found_services = []
        for service in services:
            name = service.name
            found_services.append(name)
            self.service_validators[name](service)

        self.assertEquals(5, len(found_services))
        self.assertIn('HDFS', found_services)
        self.assertIn('MAPREDUCE', found_services)
        self.assertIn('GANGLIA', found_services)
        self.assertIn('NAGIOS', found_services)
        self.assertIn('AMBARI', found_services)

    def _assert_hdfs(self, service):
        self.assertEquals('HDFS', service.name)

        found_components = {}
        for component in service.components:
            found_components[component.name] = component

        self.assertEqual(4, len(found_components))
        self._assert_component('NAMENODE', 'MASTER', "1",
                               found_components['NAMENODE'])
        self._assert_component('DATANODE', 'SLAVE', "1+",
                               found_components['DATANODE'])
        self._assert_component('SECONDARY_NAMENODE', 'MASTER', "1",
                               found_components['SECONDARY_NAMENODE'])
        self._assert_component('HDFS_CLIENT', 'CLIENT', "1+",
                               found_components['HDFS_CLIENT'])
        #TODO(jspeidel) config

    def _assert_mr(self, service):
        self.assertEquals('MAPREDUCE', service.name)

        found_components = {}
        for component in service.components:
            found_components[component.name] = component

        self.assertEqual(3, len(found_components))
        self._assert_component('JOBTRACKER', 'MASTER', "1",
                               found_components['JOBTRACKER'])
        self._assert_component('TASKTRACKER', 'SLAVE', "1+",
                               found_components['TASKTRACKER'])
        self._assert_component('MAPREDUCE_CLIENT', 'CLIENT', "1+",
                               found_components['MAPREDUCE_CLIENT'])
        # TODO(jspeidel) config

    def _assert_nagios(self, service):
        self.assertEquals('NAGIOS', service.name)

        found_components = {}
        for component in service.components:
            found_components[component.name] = component

        self.assertEqual(1, len(found_components))
        self._assert_component('NAGIOS_SERVER', 'MASTER', "1",
                               found_components['NAGIOS_SERVER'])

    def _assert_ganglia(self, service):
        self.assertEquals('GANGLIA', service.name)

        found_components = {}
        for component in service.components:
            found_components[component.name] = component

        self.assertEqual(2, len(found_components))
        self._assert_component('GANGLIA_SERVER', 'MASTER', "1",
                               found_components['GANGLIA_SERVER'])
        self._assert_component('GANGLIA_MONITOR', 'SLAVE', "1+",
                               found_components['GANGLIA_MONITOR'])

    def _assert_ambari(self, service):
        self.assertEquals('AMBARI', service.name)

        found_components = {}
        for component in service.components:
            found_components[component.name] = component

        self.assertEqual(2, len(found_components))
        self._assert_component('AMBARI_SERVER', 'MASTER', "1",
                               found_components['AMBARI_SERVER'])
        self._assert_component('AMBARI_AGENT', 'SLAVE', "1+",
                               found_components['AMBARI_AGENT'])

    def _assert_component(self, name, comp_type, cardinality, component):
        self.assertEquals(name, component.name)
        self.assertEquals(comp_type, component.type)
        self.assertEquals(cardinality, component.cardinality)

    def _assert_configurations(self, configurations):
        self.assertEquals(4, len(configurations))
        self.assertIn('global', configurations)
        self.assertIn('core-site', configurations)
        self.assertIn('mapred-site', configurations)
        self.assertIn('hdfs-site', configurations)

    def _assert_host_role_mappings(self, node_groups):
        self.assertEquals(2, len(node_groups))
        self.assertIn('MASTER', node_groups)
        self.assertIn('SLAVE', node_groups)

        master_node_group = node_groups['MASTER']
        self.assertEquals('MASTER', master_node_group.name)
        self.assertEquals(None, master_node_group.predicate)
        self.assertEquals('1', master_node_group.cardinality)
        self.assertEquals(1, master_node_group.default_count)
        self.assertEquals(8, len(master_node_group.components))
        self.assertIn('NAMENODE', master_node_group.components)
        self.assertIn('JOBTRACKER', master_node_group.components)
        self.assertIn('SECONDARY_NAMENODE', master_node_group.components)
        self.assertIn('GANGLIA_SERVER', master_node_group.components)
        self.assertIn('GANGLIA_MONITOR', master_node_group.components)
        self.assertIn('NAGIOS_SERVER', master_node_group.components)
        self.assertIn('AMBARI_SERVER', master_node_group.components)
        self.assertIn('AMBARI_AGENT', master_node_group.components)

        slave_node_group = node_groups['SLAVE']
        self.assertEquals('SLAVE', slave_node_group.name)
        self.assertEquals(None, slave_node_group.predicate)
        self.assertEquals('1+', slave_node_group.cardinality)
        self.assertEquals(2, slave_node_group.default_count)
        self.assertEquals(6, len(slave_node_group.components))
        self.assertIn('DATANODE', slave_node_group.components)
        self.assertIn('TASKTRACKER', slave_node_group.components)
        self.assertIn('GANGLIA_MONITOR', slave_node_group.components)
        self.assertIn('HDFS_CLIENT', slave_node_group.components)
        self.assertIn('MAPREDUCE_CLIENT', slave_node_group.components)
        self.assertIn('AMBARI_AGENT', slave_node_group.components)

    def setUp(self):
        self.service_validators['HDFS'] = self._assert_hdfs
        self.service_validators['MAPREDUCE'] = self._assert_mr
        self.service_validators['GANGLIA'] = self._assert_ganglia
        self.service_validators['NAGIOS'] = self._assert_nagios
        self.service_validators['AMBARI'] = self._assert_ambari


class TestServer():
    def __init__(self, hostname, role, img, flavor, public_ip, private_ip,
                 node_processes=None):
        self.hostname = hostname
        self.fqdn = hostname
        self.role = role
        self.nova_info = TestNova
        self.nova_info.image = img
        self.nova_info.flavor = flavor
        self.management_ip = public_ip
        self.public_ip = public_ip
        self.internal_ip = private_ip
        self.node_processes = node_processes


class TestNova():
    image = None
    flavor = None


class TestCluster():
    instances = []



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo.config import cfg

cli_opts = [
    cfg.StrOpt('host', default='',
               help='Hostname of IP address that will be used to listen on'),
    cfg.IntOpt('port', default=8386,
               help='Port that will be used to listen on'),
    cfg.BoolOpt('log-exchange', default=False,
                help='Log request/response exchange details: environ, '
                     'headers and bodies')
]

CONF = cfg.CONF
CONF.register_cli_opts(cli_opts)

ARGV = []


def parse_configs(argv=None, conf_files=None):
    if argv is not None:
        global ARGV
        ARGV = argv
    try:
        CONF(ARGV, project='savanna', default_config_files=conf_files)
    except cfg.RequiredOptError as roe:
        # TODO(slukjanov): replace RuntimeError with Savanna-specific exception
        raise RuntimeError("Option '%s' is required for config group "
                           "'%s'" % (roe.opt_name, roe.group.name))



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import telnetlib
import time

from savanna.tests.integration import base
import savanna.tests.integration.configs.parameters as param


@base.swift_test
class SwiftTest(base.ITestCase):

    def setUp(self):
        super(SwiftTest, self).setUp()

        telnetlib.Telnet(self.host, self.port)

    def _check_availability_swift_cluster(self, data):
        this_dir = os.getcwd()

        self.execute_command(data['namenode_ip'], 'sudo mkdir /tmp/swift && \
        sudo chmod -R 777 /tmp/swift ')

        self.write_file_to(data['namenode_ip'],
                           '/tmp/swift/generate_job_file.py',
                           open(
                               '%s/integration/swift_test/generate_job_file.py'
                               % this_dir).read())

        self.execute_command(data['namenode_ip'],
                             'python /tmp/swift/generate_job_file.py && \
                        sudo apt-get install -y python-pip && \
                        sudo pip install python-swiftclient && \
                        sudo pip install python-keystoneclient && \
                        export ST_AUTH=%s && \
                        export ST_USER=%s:admin && \
                        export ST_KEY=%s && \
                        swift -V2.0 post hadoop-job-files && \
                        swift -V2.0 upload hadoop-job-files \
                        hadoop-job-file.txt && \
                        sudo su -c "hadoop distcp -D \
                        fs.swift.service.savanna.username=%s -D \
                        fs.swift.service.savanna.password=%s \
        swift://hadoop-job-files.savanna/hadoop-job-file.txt /" hadoop && \
                        swift -V2.0 delete hadoop-job-files && \
                        hadoop dfs -copyToLocal /hadoop-job-file.txt \
                        /home/%s/hadoop-job-swift-file.txt'
                             % (param.OS_AUTH_URL,
                                param.OS_TENANT_NAME,
                                param.OS_PASSWORD,
                                param.OS_USERNAME,
                                param.OS_PASSWORD,
                                param.NODE_USERNAME))

        self.assertEquals(
            self.execute_command(
                data['namenode_ip'], 'diff hadoop-job-swift-file.txt \
                hadoop-job-file.txt')[0], 0,
            'hadoop-job-swift-file.txt != hadoop-job-file.txt')

    def test_swift(self):
        """This test checks swift work
        """
        node_processes = {'JT+NN': 1, 'TT+DN': 2}

        cluster_body = self.make_cl_body_node_processes(node_processes)
        cluster_body['name'] = param.CLUSTER_NAME_SWIFT

        try:
            cluster_id = self.create_cluster_and_get_id(cluster_body)

            instances_ip = self.get_instances_ip_and_node_processes_list(
                cluster_id)

            time.sleep(10)

            data = self.get_namenode_ip_and_tt_dn_count(instances_ip)

            self.await_active_workers_for_namenode(data)

            self._check_availability_swift_cluster(data)

        finally:
            self.del_object(self.url_cluster_with_slash, cluster_id, 204)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json


def url_for(headers, service_type, admin=False, endpoint_type=None):
    if not endpoint_type:
        endpoint_type = 'publicURL'
    if admin:
        endpoint_type = 'adminURL'

    catalog = headers['X-Service-Catalog']
    service = _get_service_from_catalog(catalog, service_type)

    if service:
        return service['endpoints'][0][endpoint_type]
    else:
        raise Exception('Service "%s" not found' % service_type)


def _get_service_from_catalog(catalog, service_type):
    if catalog:
        catalog = json.loads(catalog)
        for service in catalog:
            if service['type'] == service_type:
                return service

    return None



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from savanna.db import storage as s
from savanna.openstack.common import uuidutils
from savanna.tests.unit import base


def _create_clusters(name="cluster-1", plugin_name="some_plugin",
                     hadoop_version="1.2.3", **kwargs):
    cluster_dict = {
        "name": name,
        "plugin_name": plugin_name,
        "hadoop_version": hadoop_version,
    }
    cluster_dict.update(kwargs)

    return cluster_dict, s.create_cluster(cluster_dict)


class ClusterStorageTest(base.DbTestCase):

    def test_create_cluster_trivial(self):
        cluster_dict, cluster = _create_clusters()

        self.assertIsNotNone(cluster)
        self.assertTrue(uuidutils.is_uuid_like(cluster.id))
        self.assertDictContainsSubset(cluster_dict, cluster.dict)

    def test_clusters_multi_tenancy(self):
        self.assertEqual(0, len(s.get_clusters()))

        self.set_tenant("t-1")
        self.assertEqual(0, len(s.get_clusters()))
        _create_clusters("c-1")
        _create_clusters("c-2")
        self.assertEqual(2, len(s.get_clusters()))

        self.set_tenant("t-2")
        self.assertEqual(0, len(s.get_clusters()))
        _create_clusters("c-1")
        _create_clusters("c-2")
        self.assertEqual(2, len(s.get_clusters()))
        _create_clusters("c-3")
        self.assertEqual(3, len(s.get_clusters()))



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import contextlib
import errno
import os

from savanna.openstack.common import excutils
from savanna.openstack.common.gettextutils import _
from savanna.openstack.common import log as logging

LOG = logging.getLogger(__name__)

_FILE_CACHE = {}


def ensure_tree(path):
    """Create a directory (and any ancestor directories required)

    :param path: Directory to create
    """
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            if not os.path.isdir(path):
                raise
        else:
            raise


def read_cached_file(filename, force_reload=False):
    """Read from a file if it has been modified.

    :param force_reload: Whether to reload the file.
    :returns: A tuple with a boolean specifying if the data is fresh
              or not.
    """
    global _FILE_CACHE

    if force_reload and filename in _FILE_CACHE:
        del _FILE_CACHE[filename]

    reloaded = False
    mtime = os.path.getmtime(filename)
    cache_info = _FILE_CACHE.setdefault(filename, {})

    if not cache_info or mtime > cache_info.get('mtime', 0):
        LOG.debug(_("Reloading cached file %s") % filename)
        with open(filename) as fap:
            cache_info['data'] = fap.read()
        cache_info['mtime'] = mtime
        reloaded = True
    return (reloaded, cache_info['data'])


def delete_if_exists(path):
    """Delete a file, but ignore file not found error.

    :param path: File to delete
    """

    try:
        os.unlink(path)
    except OSError as e:
        if e.errno == errno.ENOENT:
            return
        else:
            raise


@contextlib.contextmanager
def remove_path_on_error(path):
    """Protect code that wants to operate on PATH atomically.
    Any exception will cause PATH to be removed.

    :param path: File to work with
    """
    try:
        yield
    except Exception:
        with excutils.save_and_reraise_exception():
            delete_if_exists(path)


def file_open(*args, **kwargs):
    """Open file

    see built-in file() documentation for more details

    Note: The reason this is kept in a separate module is to easily
    be able to provide a stub module that doesn't alter system
    state at all (for unit tests)
    """
    return file(*args, **kwargs)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest2

from savanna.db import models as m
from savanna.plugins.vanilla import config_helper as c_h
from savanna.plugins.vanilla import exceptions as ex
from savanna.plugins.vanilla import plugin as p


class VanillaPluginTest(unittest2.TestCase):
    def setUp(self):
        self.pl = p.VanillaProvider()
        self.cl = m.Cluster("cluster1", "tenant1", "vanilla", "1.1.2")
        self.ng1 = m.NodeGroup("nn", "f1", ["namenode"], 1)
        self.ng2 = m.NodeGroup("jt", "f1", ["jobtracker"], 1)
        self.ng3 = m.NodeGroup("tt", "f1", ["tasktracker"], 10)
        self.cl_configs = self.pl.get_configs("1.1.2")

    def test_validate(self):
        self.cl.node_groups = [self.ng1]
        self.pl.validate(self.cl)
        with self.assertRaises(ex.NotSingleNameNodeException):
            self.ng1.count = 0
            self.pl.validate(self.cl)
        with self.assertRaises(ex.NotSingleNameNodeException):
            self.ng1.count = 2
            self.pl.validate(self.cl)
        self.ng1.count = 1

        self.cl.node_groups.append(self.ng2)
        self.pl.validate(self.cl)
        with self.assertRaises(ex.NotSingleJobTrackerException):
            self.ng2.count = 2
            self.pl.validate(self.cl)

        self.cl.node_groups.append(self.ng3)
        self.ng2.count = 1
        self.pl.validate(self.cl)
        with self.assertRaises(ex.TaskTrackersWithoutJobTracker):
            self.ng2.count = 0
            self.pl.validate(self.cl)

    def test_get_configs(self):
        for cfg in self.cl_configs:
            if cfg.config_type is "bool":
                self.assertIsInstance(cfg.default_value, bool)
            elif cfg.config_type is "int":
                self.assertIsInstance(cfg.default_value, int)
            else:
                self.assertIsInstance(cfg.default_value, str)
            self.assertNotIn(cfg.name, c_h.HIDDEN_CONFS)

    def test_extract_environment_configs(self):
        env_configs = {
            "MapReduce": {
                'Job Tracker Heap Size': 1000,
                'Task Tracker Heap Size': "2000"
            },
            "HDFS": {
                'Name Node Heap Size': 3000,
                'Data Node Heap Size': "4000"
            },
            "Wrong-applicable-target": {
                't1': 4
            }}
        self.assertListEqual(c_h.extract_environment_confs(env_configs),
                             ['HADOOP_NAMENODE_OPTS=\\"-Xmx3000m\\"',
                              'HADOOP_DATANODE_OPTS=\\"-Xmx4000m\\"',
                              'HADOOP_JOBTRACKER_OPTS=\\"-Xmx1000m\\"',
                              'HADOOP_TASKTRACKER_OPTS=\\"-Xmx2000m\\"'])

    def test_extract_xml_configs(self):
        xml_configs = {
            "HDFS": {
                'dfs.replication': 3,
                'fs.default.name': 'hdfs://',
                'key': 'value'
            },
            "MapReduce": {
                'io.sort.factor': 10,
                'mapred.reduce.tasks': 2
            },
            "Wrong-applicable-target": {
                'key': 'value'
            }
        }

        self.assertListEqual(c_h.extract_xml_confs(xml_configs),
                             [('fs.default.name', 'hdfs://'),
                              ('dfs.replication', 3),
                              ('mapred.reduce.tasks', 2),
                              ('io.sort.factor', 10)])



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import savanna.openstack.common.exception as ex


class SavannaException(ex.ApiError):
    """Base Exception for the project

    To correctly use this class, inherit from it and define
    a 'message' and 'code' properties.
    """
    message = "An unknown exception occurred"
    code = "UNKNOWN_EXCEPTION"

    def __str__(self):
        return self.message


class NotFoundException(SavannaException):

    message = "Object not found"
    # It could be a various property of object which was not found
    value = None

    def __init__(self, value, message=None):
        self.code = "NOT_FOUND"
        self.value = value
        if message:
            self.message = message % value


class NameAlreadyExistsException(SavannaException):

    message = "Name already exists"

    def __init__(self, message=None):
        self.code = "NAME_ALREADY_EXISTS"
        if message:
            self.message = message


class InvalidException(SavannaException):

    message = "Invalid object reference"

    def __init__(self, message=None):
        self.code = "INVALID_REFERENCE"
        if message:
            self.message = message



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock

from savanna import context as ctx
import savanna.db.models as m
from savanna.service import instances
from savanna.tests.unit import base as models_test_base
import savanna.utils.crypto as c


class TestClusterRollBack(models_test_base.DbTestCase):
    @mock.patch('savanna.utils.openstack.nova.client')
    def test_cluster_creation_with_errors(self, novaclient):
        node_groups = [m.NodeGroup("test_group", "test_flavor",
                                   ["data node", "test tracker"], 2)]
        node_groups[0]._username = "root"

        cluster = _create_cluster_mock(node_groups, [])

        nova = _create_nova_mock(novaclient)
        nova.servers.create.side_effect = [_mock_instance(1),
                                           Exception("test")]

        nova.servers.list = mock.MagicMock(return_value=[_mock_instance(1)])

        with self.assertRaises(Exception):
            instances.create_cluster(cluster)

        session = ctx.ctx().session
        with session.begin():
            self.assertEqual(session.query(m.Instance).count(), 0)


class NodePlacementTest(models_test_base.DbTestCase):
    @mock.patch('savanna.utils.openstack.nova.client')
    def test_one_node_groups_and_one_affinity_group(self, novaclient):
        node_groups = [m.NodeGroup("test_group", "test_flavor",
                                   ["data node", "test tracker"], 2)]
        node_groups[0]._username = "root"
        cluster = _create_cluster_mock(node_groups, ["data node"])
        nova = _create_nova_mock(novaclient)
        instances._create_instances(cluster)
        userdata = _generate_user_data_script(cluster)

        nova.servers.create.assert_has_calls(
            [mock.call("test_cluster-test_group-001",
                       "initial",
                       "test_flavor",
                       scheduler_hints=None,
                       userdata=userdata,
                       key_name='user_keypair'),
             mock.call("test_cluster-test_group-002",
                       "initial",
                       "test_flavor",
                       scheduler_hints={'different_host': ["1"]},
                       userdata=userdata,
                       key_name='user_keypair')],
            any_order=False)

        session = ctx.ctx().session
        with session.begin():
            self.assertEqual(session.query(m.Instance).count(), 2)

    @mock.patch('savanna.utils.openstack.nova.client')
    def test_one_node_groups_and_no_affinity_group(self, novaclient):
        node_groups = [m.NodeGroup("test_group", "test_flavor",
                                   ["data node", "test tracker"], 2)]
        node_groups[0]._username = "root"
        cluster = _create_cluster_mock(node_groups, [])
        nova = _create_nova_mock(novaclient)
        instances._create_instances(cluster)
        userdata = _generate_user_data_script(cluster)

        nova.servers.create.assert_has_calls(
            [mock.call("test_cluster-test_group-001",
                       "initial",
                       "test_flavor",
                       scheduler_hints=None,
                       userdata=userdata,
                       key_name='user_keypair'),
             mock.call("test_cluster-test_group-002",
                       "initial",
                       "test_flavor",
                       scheduler_hints=None,
                       userdata=userdata,
                       key_name='user_keypair')],
            any_order=False)

        session = ctx.ctx().session
        with session.begin():
            self.assertEqual(session.query(m.Instance).count(), 2)

    @mock.patch('savanna.utils.openstack.nova.client')
    def test_two_node_groups_and_one_affinity_group(self, novaclient):
        node_groups = [m.NodeGroup("test_group_1", "test_flavor",
                                   ["data node", "test tracker"], 2),
                       m.NodeGroup("test_group_2", "test_flavor",
                                   ["data node", "test tracker"], 1)]
        node_groups[0]._username = "root"
        node_groups[1]._username = "root"
        cluster = _create_cluster_mock(node_groups, ["data node"])
        nova = _create_nova_mock(novaclient)
        instances._create_instances(cluster)
        userdata = _generate_user_data_script(cluster)

        nova.servers.create.assert_has_calls(
            [mock.call("test_cluster-test_group_1-001",
                       "initial",
                       "test_flavor",
                       scheduler_hints=None,
                       userdata=userdata,
                       key_name='user_keypair'),
             mock.call("test_cluster-test_group_1-002",
                       "initial",
                       "test_flavor",
                       scheduler_hints={'different_host': ["1"]},
                       userdata=userdata,
                       key_name='user_keypair'),
             mock.call("test_cluster-test_group_2-001",
                       "initial",
                       "test_flavor",
                       scheduler_hints={'different_host': ["1", "2"]},
                       userdata=userdata,
                       key_name='user_keypair')],
            any_order=False)

        session = ctx.ctx().session
        with session.begin():
            self.assertEqual(session.query(m.Instance).count(), 3)


def _create_cluster_mock(node_groups, aa):
    cluster = m.Cluster("test_cluster", "tenant_id", "mock_plugin",
                        "mock_version", "initial",
                        user_keypair_id='user_keypair', anti_affinity=aa)

    cluster._user_kp = mock.Mock()
    cluster._user_kp.public_key = "123"
    cluster.private_key = c.generate_private_key()
    cluster.node_groups = node_groups

    return cluster


def _mock_instance(id):
    instance1 = mock.Mock()
    instance1.id = id
    return instance1


def _mock_instances(count):
    return [_mock_instance(str(i)) for i in range(1, count + 1)]


def _generate_user_data_script(cluster):
    script_template = """#!/bin/bash
echo "%(public_key)s" >> %(user_home)s/.ssh/authorized_keys
echo "%(private_key)s" > %(user_home)s/.ssh/id_rsa
"""
    key = c.private_key_to_public_key(cluster.private_key)
    return script_template % {
        "public_key": key,
        "private_key": cluster.private_key,
        "user_home": "/root/"
    }


def _create_nova_mock(novalcient):
    nova = mock.Mock()
    novalcient.return_value = nova
    nova.servers.create.side_effect = _mock_instances(3)
    return nova



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import errno
import functools
import os
import shutil
import tempfile
import time
import weakref

from eventlet import semaphore
from oslo.config import cfg

from savanna.openstack.common import fileutils
from savanna.openstack.common.gettextutils import _
from savanna.openstack.common import local
from savanna.openstack.common import log as logging


LOG = logging.getLogger(__name__)


util_opts = [
    cfg.BoolOpt('disable_process_locking', default=False,
                help='Whether to disable inter-process locks'),
    cfg.StrOpt('lock_path',
               help=('Directory to use for lock files. Default to a '
                     'temp directory'))
]


CONF = cfg.CONF
CONF.register_opts(util_opts)


def set_defaults(lock_path):
    cfg.set_defaults(util_opts, lock_path=lock_path)


class _InterProcessLock(object):
    """Lock implementation which allows multiple locks, working around
    issues like bugs.debian.org/cgi-bin/bugreport.cgi?bug=632857 and does
    not require any cleanup. Since the lock is always held on a file
    descriptor rather than outside of the process, the lock gets dropped
    automatically if the process crashes, even if __exit__ is not executed.

    There are no guarantees regarding usage by multiple green threads in a
    single process here. This lock works only between processes. Exclusive
    access between local threads should be achieved using the semaphores
    in the @synchronized decorator.

    Note these locks are released when the descriptor is closed, so it's not
    safe to close the file descriptor while another green thread holds the
    lock. Just opening and closing the lock file can break synchronisation,
    so lock files must be accessed only using this abstraction.
    """

    def __init__(self, name):
        self.lockfile = None
        self.fname = name

    def __enter__(self):
        self.lockfile = open(self.fname, 'w')

        while True:
            try:
                # Using non-blocking locks since green threads are not
                # patched to deal with blocking locking calls.
                # Also upon reading the MSDN docs for locking(), it seems
                # to have a laughable 10 attempts "blocking" mechanism.
                self.trylock()
                return self
            except IOError as e:
                if e.errno in (errno.EACCES, errno.EAGAIN):
                    # external locks synchronise things like iptables
                    # updates - give it some time to prevent busy spinning
                    time.sleep(0.01)
                else:
                    raise

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.unlock()
            self.lockfile.close()
        except IOError:
            LOG.exception(_("Could not release the acquired lock `%s`"),
                          self.fname)

    def trylock(self):
        raise NotImplementedError()

    def unlock(self):
        raise NotImplementedError()


class _WindowsLock(_InterProcessLock):
    def trylock(self):
        msvcrt.locking(self.lockfile.fileno(), msvcrt.LK_NBLCK, 1)

    def unlock(self):
        msvcrt.locking(self.lockfile.fileno(), msvcrt.LK_UNLCK, 1)


class _PosixLock(_InterProcessLock):
    def trylock(self):
        fcntl.lockf(self.lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)

    def unlock(self):
        fcntl.lockf(self.lockfile, fcntl.LOCK_UN)


if os.name == 'nt':
    import msvcrt
    InterProcessLock = _WindowsLock
else:
    import fcntl
    InterProcessLock = _PosixLock

_semaphores = weakref.WeakValueDictionary()


def synchronized(name, lock_file_prefix, external=False, lock_path=None):
    """Synchronization decorator.

    Decorating a method like so::

        @synchronized('mylock')
        def foo(self, *args):
           ...

    ensures that only one thread will execute the foo method at a time.

    Different methods can share the same lock::

        @synchronized('mylock')
        def foo(self, *args):
           ...

        @synchronized('mylock')
        def bar(self, *args):
           ...

    This way only one of either foo or bar can be executing at a time.

    :param lock_file_prefix: The lock_file_prefix argument is used to provide
    lock files on disk with a meaningful prefix. The prefix should end with a
    hyphen ('-') if specified.

    :param external: The external keyword argument denotes whether this lock
    should work across multiple processes. This means that if two different
    workers both run a a method decorated with @synchronized('mylock',
    external=True), only one of them will execute at a time.

    :param lock_path: The lock_path keyword argument is used to specify a
    special location for external lock files to live. If nothing is set, then
    CONF.lock_path is used as a default.
    """

    def wrap(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            # NOTE(soren): If we ever go natively threaded, this will be racy.
            #              See http://stackoverflow.com/questions/5390569/dyn
            #              amically-allocating-and-destroying-mutexes
            sem = _semaphores.get(name, semaphore.Semaphore())
            if name not in _semaphores:
                # this check is not racy - we're already holding ref locally
                # so GC won't remove the item and there was no IO switch
                # (only valid in greenthreads)
                _semaphores[name] = sem

            with sem:
                LOG.debug(_('Got semaphore "%(lock)s" for method '
                            '"%(method)s"...'), {'lock': name,
                                                 'method': f.__name__})

                # NOTE(mikal): I know this looks odd
                if not hasattr(local.strong_store, 'locks_held'):
                    local.strong_store.locks_held = []
                local.strong_store.locks_held.append(name)

                try:
                    if external and not CONF.disable_process_locking:
                        LOG.debug(_('Attempting to grab file lock "%(lock)s" '
                                    'for method "%(method)s"...'),
                                  {'lock': name, 'method': f.__name__})
                        cleanup_dir = False

                        # We need a copy of lock_path because it is non-local
                        local_lock_path = lock_path
                        if not local_lock_path:
                            local_lock_path = CONF.lock_path

                        if not local_lock_path:
                            cleanup_dir = True
                            local_lock_path = tempfile.mkdtemp()

                        if not os.path.exists(local_lock_path):
                            fileutils.ensure_tree(local_lock_path)

                        # NOTE(mikal): the lock name cannot contain directory
                        # separators
                        safe_name = name.replace(os.sep, '_')
                        lock_file_name = '%s%s' % (lock_file_prefix, safe_name)
                        lock_file_path = os.path.join(local_lock_path,
                                                      lock_file_name)

                        try:
                            lock = InterProcessLock(lock_file_path)
                            with lock:
                                LOG.debug(_('Got file lock "%(lock)s" at '
                                            '%(path)s for method '
                                            '"%(method)s"...'),
                                          {'lock': name,
                                           'path': lock_file_path,
                                           'method': f.__name__})
                                retval = f(*args, **kwargs)
                        finally:
                            LOG.debug(_('Released file lock "%(lock)s" at '
                                        '%(path)s for method "%(method)s"...'),
                                      {'lock': name,
                                       'path': lock_file_path,
                                       'method': f.__name__})
                            # NOTE(vish): This removes the tempdir if we needed
                            #             to create one. This is used to
                            #             cleanup the locks left behind by unit
                            #             tests.
                            if cleanup_dir:
                                shutil.rmtree(local_lock_path)
                    else:
                        retval = f(*args, **kwargs)

                finally:
                    local.strong_store.locks_held.remove(name)

            return retval
        return inner
    return wrap


def synchronized_with_prefix(lock_file_prefix):
    """Partial object generator for the synchronization decorator.

    Redefine @synchronized in each project like so::

        (in nova/utils.py)
        from nova.openstack.common import lockutils

        synchronized = lockutils.synchronized_with_prefix('nova-')


        (in nova/foo.py)
        from nova import utils

        @utils.synchronized('mylock')
        def bar(self, *args):
           ...

    The lock_file_prefix argument is used to provide lock files on disk with a
    meaningful prefix. The prefix should end with a hyphen ('-') if specified.
    """

    return functools.partial(synchronized, lock_file_prefix=lock_file_prefix)
